#include "Mcl/include/CMcl.h"
#include "tclInt.h"	// for the namespace and callframe APIs
#include "cpptcl/ItclAdaptor.hpp"
#include "cpptcl/TclHash.hpp"

#include <commctrl.h>
#pragma comment (lib, "comctl32.lib")

TCL_DECLARE_MUTEX(initLock)
static int initialized = 0;


// Base class functionality for all Dialog components.
//
class DialogItem
{
    friend class DialogBox;
    friend class DialogboxItclClassExtension;

    // pack for WORD (2 byte) alignment
#   include <pshpack2.h>

    struct DLGITEMTEMPLATEEX {
	 DWORD helpID;
	 DWORD exStyle;
	 DWORD style;
	 short x;
	 short y;
	 short cx;
	 short cy;
	 WORD id;
    //   sz_Or_Ord windowClass; // name or ordinal of a window class
    //   sz_Or_Ord title;       // title string or ordinal of a resource
    //   WORD   extraCount;     // bytes of following creation data
    };

    // Resume normal packing
#   include <poppack.h>

    LPBYTE buffer;
    size_t space;
    size_t length;	// number of BYTEs
    DialogItem **us;

protected:

    virtual void OnCreation(HWND hControl) = 0;

    // ----------------------------------------------------------------------
    // Copy constructor.  Used when a DialogBox grabs us.
    // ----------------------------------------------------------------------
    DialogItem (DialogItem &other)
    {
	buffer = new BYTE [other.length];
	memcpy(buffer, other.buffer, other.length);
	us = (DialogItem **) buffer + ((LPBYTE)other.us - other.buffer);
	*us = this;
    }

    DialogItem (const wchar_t *title, UINT sty, const wchar_t *classname,
	    DWORD flags, WORD id, DWORD helpID, short x, short y, short w,
	    short h)
	: space(0), buffer(0L)
    {
	LPWORD lpw;
	DLGITEMTEMPLATEEX *lpditex;

	// Discover the buffer size.
	space = sizeof(DLGITEMTEMPLATEEX);
	if (title != 0L) {
	    space += (sizeof(wchar_t) * wcslen(title));
	} else {
	    space += 2;
	}
	if (classname != 0L) {
	    space += (sizeof(wchar_t) * wcslen(classname));
	} else {
	    space += 4;
	}
	// Creation data plus a fudge factor for alignments.
	space += sizeof(DialogItem **) + 8;

	lpw = (LPWORD) (buffer = new BYTE [space]);
	ZeroMemory(buffer, space);
	lpditex = (DLGITEMTEMPLATEEX *) lpw;
	lpditex->helpID = helpID;
	lpditex->exStyle = 0;
	lpditex->style = WS_CHILD | WS_VISIBLE | flags;
	lpditex->x  = x;
	lpditex->y  = y;
	lpditex->cx = w;
	lpditex->cy = h;
	lpditex->id = id;

	// Align the class data on a DWORD boundary.  M$ says creation data
	// needs the alignment, but I found this was wrong and alignment
	// goes here, instead.
	//
	lpw = lpwAlign((LPWORD)(lpditex + 1));

	if (classname != 0L) {
	    // wide string copy, include the terminator in the amount
	    // of WORDS copied.
	    lpw += swprintf((LPWSTR) lpw, L"%s", classname) + 1;
	} else {
	    *(lpw++) = 0xFFFF;
	    *(lpw++) = sty;
	}

	if (title != 0L) {
	    // wide string copy, include the terminator in the amount
	    // of WORDS copied.
	    lpw += swprintf((LPWSTR) lpw, L"%s", title) + 1;
	} else {
	    // no title
	    *(lpw++) = 0;
	}

	// Set creation data.  Is this 64-bit safe?
	*(lpw++) = sizeof(DialogItem **);
	us = (DialogItem **) lpw;
	*us = this;
	lpw += sizeof(DialogItem **) / sizeof(WORD);

	// Align DLGITEMTEMPLATEEX on DWORD boundary
	lpw = lpwAlign(lpw);

	// set memory length used (in bytes).
	length = ((LPBYTE)lpw - buffer);

	if (length > space) throw;
    }

    virtual ~DialogItem ()
    {
	delete [] buffer;
    }

    // ----------------------------------------------------------------------
    // Aligns a pointer to a DWORD (4 byte) boundary.
    // ----------------------------------------------------------------------
    static LPWORD lpwAlign (LPWORD lpIn)
    {
	DWORD_PTR ul;
	ul = (DWORD_PTR) lpIn;
	ul += 3;
	ul >>= 2;
	ul <<= 2;
	return (LPWORD) ul;
    }
};

// ---------  Standard system controls  -------------

class Button_DlgItem : public DialogItem
{
public:
    Button_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem(title,
	    0x0080,
	    0L,
	    BS_PUSHBUTTON | WS_TABSTOP,
	    id, helpID,	x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class DefButton_DlgItem : protected DialogItem
{
public:
    DefButton_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0080,
	    0L,
	    BS_DEFPUSHBUTTON | WS_TABSTOP,
	    id, helpID,	x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class GroupBox_DlgItem : protected DialogItem
{
public:
    GroupBox_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0080,
	    0L,
	    BS_GROUPBOX,
	    id, helpID,	x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class Edit_DlgItem : protected DialogItem
{
public:
    Edit_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0081,
	    0L,
	    WS_BORDER | WS_TABSTOP,
	    id, helpID, x, y, w, h)
    {
    }
    //ES_AUTOHSCROLL
    virtual void OnCreation(HWND hControl) {}
};
#define Label_DlgItem	Static_DlgItem
class Static_DlgItem : protected DialogItem
{
public:
    Static_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0082,
	    0L,
	    SS_NOPREFIX,
	    id, helpID, x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class ListBox_DlgItem : protected DialogItem
{
public:
    ListBox_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0083,
	    0L,
	    0,
	    id, helpID, x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class ScrollBar_DlgItem : protected DialogItem
{
public:
    ScrollBar_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0084,
	    0L,
	    0,
	    id, helpID, x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class ComboBox_DlgItem : protected DialogItem
{
public:
    ComboBox_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0085,
	    0L,
	    0,
	    id, helpID, x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class Check_DlgItem : protected DialogItem
{
    Check_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0080,
	    0L,
	    BS_AUTOCHECKBOX | WS_TABSTOP,
	    id, helpID,	x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class Radio_DlgItem : protected DialogItem
{
    Radio_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0x0080,
	    0L,
	    BS_PUSHBUTTON | WS_TABSTOP,
	    id, helpID,	x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};


// ---------  v4.71 custom controls from comctl32.dll  ----------

class Header_DlgItem : protected DialogItem
{
    Header_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0,
	    WC_HEADERW,
	    0,
	    id, helpID,	x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class ToolBar_DlgItem : protected DialogItem
{
    ToolBar_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0,
	    TOOLBARCLASSNAMEW,
	    0,
	    id, helpID,	x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class ReBar_DlgItem : protected DialogItem
{
    ReBar_DlgItem (const wchar_t *title, WORD id, DWORD helpID, short x,
	    short y, short w, short h)
	: DialogItem (title,
	    0,
	    REBARCLASSNAMEW,
	    0,
	    id, helpID,	x, y, w, h)
    {
    }
    virtual void OnCreation(HWND hControl) {}
};
class ToolTips_DlgItem : protected DialogItem
{
    // window class name is: TOOLTIPS_CLASSW
    virtual void OnCreation(HWND hControl) {}
};
class Status_DlgItem : protected DialogItem
{
    // window class name is: STATUSCLASSNAMEW
    virtual void OnCreation(HWND hControl) {}
};
class TrackBar_DlgItem : protected DialogItem
{
    // window class name is: TRACKBAR_CLASSW
    virtual void OnCreation(HWND hControl) {}
};
class UpDown_DlgItem : protected DialogItem
{
    // window class name is: UPDOWN_CLASSW
    virtual void OnCreation(HWND hControl) {}
};
class Progress_DlgItem : protected DialogItem
{
    // window class name is: PROGRESS_CLASSW
    virtual void OnCreation(HWND hControl) {}
};
class ListView_DlgItem : protected DialogItem
{
    // window class name is: WC_LISTVIEWW
    virtual void OnCreation(HWND hControl) {}
};
class TreeView_DlgItem : protected DialogItem
{
    // window class name is: WC_TREEVIEWW
    virtual void OnCreation(HWND hControl) {}
};
class ComboBoxEx_DlgItem : protected DialogItem
{
    // window class name is: WC_COMBOBOXEXW
    virtual void OnCreation(HWND hControl) {}
};
class TabCtrl_DlgItem : protected DialogItem
{
    // window class name is: WC_TABCONTROLW
    virtual void OnCreation(HWND hControl) {}
};
class Animate_DlgItem : protected DialogItem
{
    // window class name is: ANIMATE_CLASSW
    virtual void OnCreation(HWND hControl) {}
};
class MonthCal_DlgItem : protected DialogItem
{
    // window class name is: MONTHCAL_CLASSW
    virtual void OnCreation(HWND hControl) {}
};
class DateTimePick_DlgItem : protected DialogItem
{
    // window class name is: DATETIMEPICK_CLASSW
    virtual void OnCreation(HWND hControl) {}
};
class IPAddress_DlgItem : protected DialogItem
{
    // window class name is: WC_IPADDRESSW
    virtual void OnCreation(HWND hControl) {}
};
class PageScroller_DlgItem : protected DialogItem
{
    // window class name is: WC_PAGESCROLLERW
    virtual void OnCreation(HWND hControl) {}
};
class NativeFont_DlgItem : protected DialogItem
{
    // window class name is: WC_NATIVEFONTCTLW
    virtual void OnCreation(HWND hControl) {}
};
class Link_DlgItem : protected DialogItem
{
    // window class name is: WC_LINK
    virtual void OnCreation(HWND hControl) {}
};

// remove the Win API one that points to DialogBoxA()
#undef DialogBox

// Top-level class component for creating and running the dialog
//
class DialogBox
{
    // pack for WORD (2 byte) alignment
#   include <pshpack2.h>

    struct DLGTEMPLATEEX {
	 WORD   dlgVer;
	 WORD   signature;
	 DWORD  helpID;
	 DWORD  exStyle;
	 DWORD  style;
	 WORD   cdit;
	 short  x;
	 short  y;
	 short  cx;
	 short  cy;
    //     sz_Or_Ord menu;	    // name or ordinal of a menu resource
    //     sz_Or_Ord windowClass;   // name or ordinal of a window class
    //     WCHAR title[titleLen];   // title string of the dialog box
    //     WORD pointsize;	    // only if DS_SETFONT or DS_SHELLFONT flag is set
    //     WORD weight;		    // only if DS_SETFONT or DS_SHELLFONT flag is set
    //     BYTE italic;		    // only if DS_SETFONT or DS_SHELLFONT flag is set
    //     BYTE charset;	    // only if DS_SETFONT or DS_SHELLFONT flag is set
    //     WCHAR typeface[stringLen]; // only if DS_SETFONT or DS_SHELLFONT is set
    };

    // Resume normal packing
#   include <poppack.h>

    LPBYTE buffer;
    size_t space;

public:

    DialogBox (const wchar_t *title, short w, short h, DWORD helpID, DialogItem *...)
	: space(0)
    {
	LPWORD lpw;
	LPWSTR lpwsz;
	int nchar;
	DLGTEMPLATEEX *lpdt;

	// TODO: fill an array of DialogItems.

	// Discover the buffer size for the template header.
	space += sizeof(DLGTEMPLATEEX);
	space += (sizeof(wchar_t) * wcslen(title));
	// font data
	space += 6 + (sizeof(wchar_t) * wcslen(L"MS Sans Serif"));
	// fudge factor for possible alignments along with creation data.
	space += (4 + sizeof(DialogBox **));

	lpw = (LPWORD) (buffer = new BYTE [space]);
	ZeroMemory(buffer, space);

	lpdt = (DLGTEMPLATEEX *) lpw;	// Define a dialog box template.
	lpdt->dlgVer = 1;		// this is always 1.
	lpdt->signature = 0xFFFF;	// dialog template is an extended type.
	lpdt->helpID = helpID;		// help context identifier.
	lpdt->exStyle = 0;		// extended styles
	// normal styles
	lpdt->style = WS_VISIBLE | WS_POPUP | WS_SYSMENU | WS_CAPTION | DS_CENTER | 
		DS_MODALFRAME | DS_SETFOREGROUND | DS_NOFAILCREATE | DS_SETFONT | 
		DS_CONTEXTHELP;
	lpdt->cdit = 1;		// number of controls
	lpdt->x  = 0;
	lpdt->y  = 0;
	lpdt->cx = w;
	lpdt->cy = h;
	lpw = (LPWORD) (lpdt + 1);  // an alignment might be needed here -- we'll see.

	*(lpw++) = 0;		    // menu
	*(lpw++) = 0;		    // dialog box class

	if (title != 0L && *title != L'\0') {
	    // wide string copy, include the terminator in the amount
	    // of WORDS copied.
	    lpw += (swprintf((LPWSTR) lpw, L"%s", title) + 1);
	} else {
	    // no title
	    *(lpw++) = 0;
	}

	*(lpw++) = 8;                 // set font pitch
	*(lpw++) = FW_NORMAL;         // set font weight
	*(lpw++) = FALSE;             // set italic
	lpwsz = (LPWSTR) lpw;         // set font name
	nchar = MultiByteToWideChar(CP_ACP, 0, "MS Sans Serif", -1, lpwsz, 25);
	lpw   += nchar;
	lpw   = lpwAlign (lpw);  // align next DLGITEMTEMPLATEEX on a DWORD boundary
	
	//return memory length used by the header in sizeof(WORD).
	//return (lpw - lpstart);
    }

    // Aligns a pointer to a DWORD (4 byte) boundary.
    //
    static LPWORD lpwAlign (LPWORD lpIn)
    {
	DWORD_PTR ul;
	ul = (DWORD_PTR) lpIn;
	ul += 3;
	ul >>= 2;
	ul <<= 2;
	return (LPWORD) ul;
    }

    static BOOL CALLBACK
    DialogProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
	BOOL bRet = FALSE;

	switch (uMsg) {
	case WM_INITDIALOG:
	    SetWindowLongPtr(hDlg, DWLP_USER, lParam);
	    bRet = TRUE;
	    break;
//	case WM_ENTERIDLE:
//	    break;
	case WM_COMMAND:
	    switch (LOWORD(wParam)) {
	    case IDOK:
	    // fall through
	    case IDCANCEL:
		EndDialog(hDlg, uMsg);
	    }  
	    break;
	}
	return bRet;
    }
};


// --------------  Tcl (Itcl) stuff starts here ---------------


// ----------------------------------------------------------------------
// This class manages the Itcl interface.
// ----------------------------------------------------------------------
class DialogboxItclClassExtension
    : protected Itcl::IAdapter<DialogboxItclClassExtension>
{
    // allows DoCleanup() to be private.
    friend class Tcl::Adapter<DialogboxItclClassExtension>;

    // This structure is what gets queued into Tcl's event loop
    struct DIALOGEVENT {
	Tcl_Event header;
	Tcl_Obj *script;
	DialogboxItclClassExtension *ext;
    };

    // We store this in the hash table
//    struct DIALOGBOXINFO {
//      CMclThread *thread;
//      DialogBoxDriver *dlgBoxDrv;
//    };

    Tcl::Hash<DialogItem *> dlgItems;	// All our dialog components.
    Tcl::Hash<DialogBox *> dialogs;	// All our dialogs.
    Tcl_AsyncHandler Async;		// interp specific async handler.

public:

    DialogboxItclClassExtension(Tcl_Interp *interp)
	: Itcl::IAdapter<DialogboxItclClassExtension>(interp)
    {
	NewItclCmd(interp, "winutils-dialog-construct", DialogConstructCmd);
	NewItclCmd(interp, "winutils-dialog-destruct", DialogDestructCmd);
	NewItclCmd(interp, "winutils-dlg-button-construct", ButtonConstructCmd);
	NewItclCmd(interp, "winutils-dlg-button-destruct", ButtonDestructCmd);
	NewItclCmd(interp, "winutils-dlg-combo-construct", ComboConstructCmd);
	NewItclCmd(interp, "winutils-dlg-combo-destruct", ComboDestructCmd);
	NewItclCmd(interp, "winutils-dlg-label-construct", LabelConstructCmd);
	NewItclCmd(interp, "winutils-dlg-label-destruct", LabelDestructCmd);
	NewItclCmd(interp, "winutils-dlg-edit-construct", EditConstructCmd);
	NewItclCmd(interp, "winutils-dlg-edit-destruct", EditDestructCmd);
	NewItclCmd(interp, "winutils-dlg-check-construct", CheckConstructCmd);
	NewItclCmd(interp, "winutils-dlg-check-destruct", CheckDestructCmd);

	Async = Tcl_AsyncCreate(FlushDialog, this);
    }

    ~DialogboxItclClassExtension()
    {
	int result;
	DialogItem *dlgItem;
	DialogBox *dlgBox;

        for (
	    result = dlgItems.Top(&dlgItem);
	    result == TCL_OK;
	    result = dlgItems.Next(&dlgItem))
	{
	    delete dlgItem;
	}

        for (
	    result = dialogs.Top(&dlgBox);
	    result == TCL_OK;
	    result = dialogs.Next(&dlgBox))
	{
	    delete dlgBox;
	}

	// Now that all the XXXX instances that had a reference to
	// the Async token are closed, close the Async token.
	//
	Tcl_AsyncDelete(Async);

	// Now remove any events currently in the event loop that call into us.
	// This is very important because the Tcl_Obj* for the script is
	// now a dangling reference and most likely has been returned to the
	// allocator since the HotKeyListener instance gave up ownership to
	// the Tcl_Obj* when it was closed.
	//
	Tcl_DeleteEvents(HoseDialogEvents, this);
    }

private:

    // ------------------------------------------------------------------
    // This is called by the adapter base to let us know we are about to go
    // away and it's safe to use the interp for any needed cleanup.
    // ------------------------------------------------------------------
    virtual void DoCleanup() {}

    int DialogConstructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }
    int DialogDestructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	ItclObject *DialogItclObj;

	if (GetItclObj(&DialogItclObj, objv[0]) != TCL_OK) return TCL_ERROR;
	//if (dialogs.Extract(DialogItclObj, &) == TCL_OK) {
	//    delete button;
	//}
	return TCL_OK;
    }

    int ButtonConstructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	ItclObject *ButtonItclObj;
	DialogItem *button;
	Tcl_Obj *resultObj = Tcl_GetObjResult(interp);
	int i, tmp;
	short dims[4];
	const char *names[] = {"<x-origin>", "<y-origin>", "<height>", "<width>"};

	if (objc != 6) {
	    Tcl_WrongNumArgs(interp, 1, objv,
		    "<text> <x-origin> <y-origin> <height> <width>");
	}

	if (GetItclObj(&ButtonItclObj, objv[0]) != TCL_OK) return TCL_ERROR;

	for (i = 0; i < 4; i++) {
	    if (Tcl_GetIntFromObj(interp, objv[i+2], &tmp) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	    if (tmp >= 0 && tmp <= MAXSHORT) {
		dims[i] = (short) tmp;
	    } else {
		Tcl_AppendStringsToObj(resultObj, names[i],
			" was out-of-range at ", Tcl_GetString(objv[2]),
			".  Should be zero through 32,767 inclusive", 0L);
		return TCL_ERROR;
	    }
	}

	button = new Button_DlgItem(Tcl_GetUnicode(objv[1]), 20, 0, dims[0],
		dims[1], dims[2], dims[3]);

	dlgItems.Add(ButtonItclObj, button);
	return TCL_OK;
    }
    int ButtonDestructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	ItclObject *ButtonItclObj;
	DialogItem *button;

	if (GetItclObj(&ButtonItclObj, objv[0]) != TCL_OK) return TCL_ERROR;
	if (dlgItems.Extract(ButtonItclObj, &button) == TCL_OK) {
	    delete button;
	}
	return TCL_OK;
    }

    int ComboConstructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }
    int ComboDestructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }

    int LabelConstructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }
    int LabelDestructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }

    int EditConstructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }
    int EditDestructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }

    int CheckConstructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }
    int CheckDestructCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	return TCL_OK;
    }

    void QueueAllPendingDialogs () {}
    void EvalOneDialog (Tcl_Obj *script) {}

    // ------------------------------------------------------------------
    // The Tcl_AsyncProc callback.
    // ------------------------------------------------------------------
    static int FlushDialog (ClientData clientData, Tcl_Interp *, int code)
    {
	((DialogboxItclClassExtension *)clientData)->QueueAllPendingDialogs();
	return code;
    }

    // ------------------------------------------------------------------
    // The Tcl_EventProc callback.
    // ------------------------------------------------------------------
    static int EvalCallback (Tcl_Event *evPtr, int flags)
    {
	DIALOGEVENT *stuff = (DIALOGEVENT *) evPtr;

	// we only handle file-type events here.
	if (!(flags & TCL_FILE_EVENTS)) return 0;
	(stuff->ext)->EvalOneDialog(stuff->script);
	return 1;
    }

    // ----------------------------------------------------------------------
    // The Tcl_EventDeleteProc callback.
    // ----------------------------------------------------------------------
    static int HoseDialogEvents (Tcl_Event *evPtr, ClientData clientData)
    {
	// No dynamic storage was used in the structure, so we don't
	// need to do anything special in here.
	return (evPtr->proc == EvalCallback && 
	    (DialogboxItclClassExtension *)clientData == ((DIALOGEVENT *)evPtr)->ext);
    }
};


extern "C" int
Dialogbox_Init (Tcl_Interp *interp)
{
#ifdef USE_ITCL_STUBS
    if (Itcl_InitStubs(interp, "3.1", 0) == 0L) {
	return TCL_ERROR;
    }
#endif

    Tcl_MutexLock(&initLock);
    if (!initialized) {
	initialized = 1;
	INITCOMMONCONTROLSEX icce;
	BOOL rtn;
	icce.dwSize = sizeof(INITCOMMONCONTROLSEX);
	icce.dwICC = ICC_WIN95_CLASSES | ICC_DATE_CLASSES | ICC_USEREX_CLASSES |
		ICC_COOL_CLASSES;
	// assume v4.71  TODO: We should verify this.
	rtn = InitCommonControlsEx(&icce);
    }
    Tcl_MutexUnlock(&initLock);

    new DialogboxItclClassExtension(interp);
    return TCL_OK;
}
