/* ----------------------------------------------------------------------------
 * DirWatch.cpp --
 *
 *	Watches directories for changes through FindFirstChangeNotification().
 *	Fires a script when the action happens.
 *
 * ----------------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this.  See license.txt
 * for more detail.
 *
 * ----------------------------------------------------------------------------
 * RCS: @(#) $Id: DirWatch.cpp,v 1.4 2003/09/08 16:39:05 davygrvy Exp $
 * ----------------------------------------------------------------------------
 */

#ifdef _MSC_VER
#   pragma comment (lib, "user32.lib")
#endif

#include "Mcl/include/CMcl.h"
#include <ERRNO.H>

// for the namespace and callframe APIs
//#include "tclInt.h"

#include "cpptcl/ItclAdaptor.hpp"
#include "cpptcl/TclHash.hpp"


// ----------------------------------------------------------------------
// Format system error messages.
// ----------------------------------------------------------------------
static Tcl_Obj *
GetSysMsgObj (DWORD id)
{
    int chars;
    char sysMsgSpace[512];

    chars = wsprintf(sysMsgSpace, "[%u] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS|
	    FORMAT_MESSAGE_MAX_WIDTH_MASK, 0L, id, 0, sysMsgSpace,
	    (512-chars), 0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}


// ----------------------------------------------------------------------
//  This is the class we run in a thread for doing the actual watching.
// ----------------------------------------------------------------------
class DirWatcher : public CMclThreadHandler
{
public:
    DirWatcher (
	Tcl_Obj *_path,
	Tcl_Obj *_script,
	CMclQueue<Tcl_Obj *> &_workQ,
	CMclEvent &_stop,
	Tcl_AsyncHandler _async,
	CMclEvent &_ready)
	: path(Tcl_GetString(_path)),
	script(_script),
	workQ(_workQ),
	stop(_stop),
	async(_async),
	ready(_ready),
	status(NO_ERROR)
    {
	Tcl_IncrRefCount(script);
    }

    ~DirWatcher()
    {
	Tcl_DecrRefCount(script);
    }

    DWORD Status ()
    {
	return status;
    }

protected:
    unsigned ThreadHandlerProc(void)
    {
	CMclWaitableCollection Events;
	HANDLE hWatch;

	hWatch = ::FindFirstChangeNotification(path, FALSE,
	    FILE_NOTIFY_CHANGE_LAST_WRITE);
	
	if (!CMclIsValidHandle(hWatch)) {
	    status = GetLastError();
	    ready.Set();
	    return 1;
	}

	// indicate to the parent we are go.
	ready.Set();

	Events.AddObject(stop);   // index = 0  more important events are first
	Events.AddObject(hWatch); // index = 1
	
	while (1) {
	    DWORD dwEvent;
	    
	    dwEvent = Events.Wait(FALSE, INFINITE);
	    
	    if (!CMclWaitSucceeded(dwEvent, Events.GetCount()) ||
		CMclWaitSucceededIndex(dwEvent) == 0) {
		// stop event, process exiting, or some other error.
		::FindCloseChangeNotification(hWatch);
		return 0;
	    }
	    ::FindNextChangeNotification(hWatch);
	    
	    workQ.Put(script);	    // put the copy onto our work queue.
	    Tcl_AsyncMark(async);   // notify Tcl that we have work ready.
	}
	
	// just to prevent a compiler error.
	// the code never returns from here.
	return 0;
    }

private:
    const char *path;
    CMclEvent &stop;
    CMclEvent &ready;
    Tcl_Obj *script;
    CMclQueue<Tcl_Obj *> &workQ;
    Tcl_AsyncHandler async;
    DWORD status;
};


// ----------------------------------------------------------------------
// This class manages the Tcl interface.
// ----------------------------------------------------------------------
class DirwatchExtension : public Itcl::IAdaptor<DirwatchExtension>
{
    // This structure is what gets queued into Tcl's event loop
    struct DIRWATCHEVENT {
	Tcl_Event header;
	Tcl_Obj *script;
	DirwatchExtension *ext;
    };

    // We store this in the hash table
    struct DIRWATCHINFO {
      CMclThread *thread;
      CMclEvent *stop;
      DirWatcher *watcher;
    };

    Tcl::Hash<DIRWATCHINFO *> watches; // All our hotkeys in this interp.
    Tcl_AsyncHandler Async;	    // interp specific async handler.
    CMclQueue<Tcl_Obj *> workQ;	    // the work queue for jobs.
    unsigned int watchCnt;	    // Used for the token ID.

public:
    DirwatchExtension(Tcl_Interp *interp)
	: Itcl::IAdaptor<DirwatchExtension>(interp), watchCnt(0)
    {
	NewItclCmd("winutils-dirwatch-construct", DirWatchConstructCmd);
	NewItclCmd("winutils-dirwatch-destruct", DirWatchDestructCmd);
	Async = Tcl_AsyncCreate(FlushWatches, this);
    }
    ~DirwatchExtension()
    {
	// We are NOT allowed to use the interp pointer in here.
	//
	DIRWATCHINFO *dwInfo;
	int result;

        for (
	    result = watches.Top(&dwInfo);
	    result == TCL_OK;
	    result = watches.Next(&dwInfo)
        ) {
	    dwInfo->stop->Set();

	    // The thread handle is placed in the signaled state when it
	    // finishes.  Use a timeout just to be safe to avoid a lock-up.
	    //
	    dwInfo->thread->Wait(500);

	    delete dwInfo->thread;
	    delete dwInfo->stop;
	    delete dwInfo->watcher;
	    delete dwInfo;
	}

	// Now that all the HotKeyListener instances that had a reference to
	// the Async token are closed, close the Async token.
	//
	Tcl_AsyncDelete(Async);

	// Now remove any events currently in the event loop that call into us.
	// This is very important because the Tcl_Obj* for the script is
	// now a dangling reference and most likely has been returned to the
	// allocator since the HotKeyListener instance gave up ownership to
	// the Tcl_Obj* when it was closed.
	//
	Tcl_DeleteEvents(HoseWatchEvents, this);
    }

private:

    // ------------------------------------------------------------------
    // Make a new one and return a token.
    // ------------------------------------------------------------------
    int DirWatchConstructCmd (int objc, struct Tcl_Obj * const objv[])
    {
	DIRWATCHINFO *dwInfo;
	CMclEvent ready;
	DWORD err;
	char buff[40];

	if (objc != 4) {
	    Tcl_WrongNumArgs(interp, 2, objv, "<directory> <script>");
	    return TCL_ERROR;
	}

	dwInfo = new DIRWATCHINFO;
	dwInfo->stop = new CMclEvent();
	dwInfo->watcher = new DirWatcher(objv[2], objv[3], workQ, *dwInfo->stop, 
		Async, ready);

	// The new thread is created here.  Use a tiny stack size.
	//
	dwInfo->thread = new CMclThread(dwInfo->watcher, 0, 0L, 256);

	if ((err = dwInfo->thread->Status()) != NO_ERROR) {
	    Tcl_Obj *msg;

	    switch (err) {
	    case EAGAIN:
		msg = Tcl_NewStringObj("Too many threads open.  Can't create any more.",
			(size_t) -1);
		break;
	    case EINVAL:
		msg = Tcl_NewStringObj("Bad argument.", (size_t) -1);
		break;
	    default:
		msg = GetSysMsgObj(err);
	    }
       	    Tcl_SetObjResult(interp, msg);
	    delete dwInfo->stop;
	    delete dwInfo->watcher;
	    delete dwInfo->thread;
	    delete dwInfo;
	    return TCL_ERROR;
	}

	// Block until after it calls FindFirstChangeNotification(), and
	// we know if it failed.
	//
	ready.Wait(INFINITE);
	if ((err = dwInfo->watcher->Status()) != NO_ERROR) {
	    Tcl_SetObjResult(interp, GetSysMsgObj(err));
	    delete dwInfo->stop;
	    delete dwInfo->thread;
	    delete dwInfo->watcher;
	    delete dwInfo;
	    return TCL_ERROR;
	}

	// save the new HOTKEYINFO structure into the hash table associated
	// to the token.
	//
	wsprintf(buff, "watch%u", watchCnt++);
	watches.Add(buff, dwInfo);
	Tcl_SetObjResult(interp, Tcl_NewStringObj(buff, -1));
	return TCL_OK;
    }

    // ------------------------------------------------------------------
    // Delete one from the token.
    // ------------------------------------------------------------------
    int DirWatchDestructCmd (int objc, struct Tcl_Obj * const objv[])
    {
	DIRWATCHINFO *dwInfo;

	if (watches.Extract(Tcl_GetString(objv[2]), &dwInfo) == TCL_OK) {
	    dwInfo->stop->Set();

	    // The thread handle is placed in the signaled state when it
	    // finishes.  Use a timeout just to be extra safe to avoid a
	    // lock-up.
	    //
	    dwInfo->thread->Wait(500);

	    delete dwInfo->thread;
	    delete dwInfo->stop;
	    delete dwInfo->watcher;
	    delete dwInfo;

	    // Possible bug where if an event is queued, but we can't
	    // delete it from here exists!

	} else {
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp), "\"",
		    Tcl_GetString(objv[2]), "\" was not a valid token.", 0L);
	    return TCL_ERROR;
	}
	return TCL_OK;
    }

    // ------------------------------------------------------------------
    // Our Tcl_AsyncProc is calling back into us.
    // ------------------------------------------------------------------
    void QueueAllPendingWatches ()
    {
	Tcl_Obj *script;
	DIRWATCHEVENT *event;

	// WARNING: Only Tcl allocation routines are allowed in here.

	// Some running HotKeyListener instance called Tcl_AsyncMark.  There
	// must be at least one entry in the WorkQueue, but pull them all off
	// while we're here.  Don't let a good context go to waste.
	//
	while (workQ.Get(script, 0)) {

	    event = (DIRWATCHEVENT *) ckalloc(sizeof(DIRWATCHEVENT));
	    event->header.proc = EvalCallback;
	    event->ext    = this;
	    event->script = script;   // you came from a distant thread

	    // Queue it into Tcl's event loop. [like Win32's PostMessage()]
	    Tcl_QueueEvent((Tcl_Event *)event, TCL_QUEUE_TAIL);
	}
    }

    // ------------------------------------------------------------------
    // Our Tcl_EventProc is calling back into us from the event loop.
    // ------------------------------------------------------------------
    void EvalOneWatch (Tcl_Obj *script)
    {
	CallFrame frame;
	Proc proc;
	Tcl_Namespace *nsPtr;
	int result;

	// let's have some fun and run it like a proc AND within a namespace
	nsPtr = Tcl_FindNamespace(interp, "::winutils", 0L, 0);
	Tcl_PushCallFrame(interp, (Tcl_CallFrame *)&frame, nsPtr, 1);

	proc.numCompiledLocals = 0;
	proc.firstLocalPtr = 0;
	proc.lastLocalPtr = 0;
	frame.procPtr = &proc;

	result = Tcl_EvalObj(interp, script);
	Tcl_PopCallFrame(interp);

	if (result == TCL_ERROR) {
	    Tcl_Obj *buff;
	    int len;
	    const char *str;

	    buff = Tcl_NewStringObj("\nError in watch script \"", -1);
	    Tcl_AppendObjToObj(buff, script);
	    Tcl_AppendStringsToObj(buff, "\"", 0L);
	    str = Tcl_GetStringFromObj(buff, &len);
	    Tcl_AddObjErrorInfo(interp, str, len);
	    Tcl_DecrRefCount(buff);
	    Tcl_BackgroundError(interp);
	}

	Tcl_ResetResult(interp);
    }

    // ----------------------------------------------------------------------
    // The Tcl_AsyncProc callback.
    // ----------------------------------------------------------------------
    static int
    FlushWatches (ClientData clientData, Tcl_Interp *, int code)
    {
	((DirwatchExtension *)clientData)->QueueAllPendingWatches();
	return code;
    }

    // ----------------------------------------------------------------------
    // The Tcl_EventProc callback.
    // ----------------------------------------------------------------------
    static int
    EvalCallback (Tcl_Event *evPtr, int flags)
    {
	DIRWATCHEVENT *event = reinterpret_cast<DIRWATCHEVENT *>(evPtr);

	// we only handle file-type events here.
	if (!(flags & TCL_FILE_EVENTS)) return 0;
	(event->ext)->EvalOneWatch(event->script);
	return 1;
    }

    // ----------------------------------------------------------------------
    // The Tcl_EventDeleteProc callback.
    // ----------------------------------------------------------------------
    static int
    HoseWatchEvents (Tcl_Event *evPtr, ClientData clientData)
    {
	// No dynamic storage was used in the structure, so we don't
	// need to do anything special in here.
	return (evPtr->proc == EvalCallback && 
	    (DirwatchExtension *)clientData == ((DIRWATCHEVENT *)evPtr)->ext);
    }
};


// ----------------------------------------------------------------------
// The Tcl_PackageInitProc for this section.
// ----------------------------------------------------------------------
extern "C" int
Dirwatch_Init (Tcl_Interp *interp)
{
#ifdef USE_ITCL_STUBS
    if (Itcl_InitStubs(interp, "3.1", 0) == 0L) {
	return TCL_ERROR;
    }
#else
    if (Itcl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
#endif
    new DirwatchExtension(interp);
    return TCL_OK;
}
