/*************************************************************
 ** KeyLock.c  --  (was) part of the Falcon DLL Extension v1.60
 **
 ** Abstract: Gets/Sets the NumLock, CapsLock, and ScrollLock
 **           states of the keyboard.
 **
 ** Author: David Gravereaux  mailto:davygrvy@pobox.com
 ************************************************************/

#include "tcl.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#ifdef _MSC_VER
#   pragma comment (lib, "user32.lib")
#endif


/* return the toggle state. */
#define GetLock(key)  (GetKeyState(key) & 0x01)

/* Flip the toggle state. */
#define TogLock(key) \
    keybd_event(key, 0x45, KEYEVENTF_EXTENDEDKEY | 0, 0); \
    keybd_event(key, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0)


/*
 *----------------------------------------------------------------------
 *
 * KeyLockObjCmd --
 *
 *  This function implements the Tcl "winutils::keylock" command.
 *
 * Results:
 *  1 or 0 for a get, or nothing for a set.
 *
 * Side effects:
 *  It might not be compatible for Win98 and some
 *  NT keyboard drivers.
 *
 *----------------------------------------------------------------------
 */
static int
KeyLockObjCmd (ClientData clientData, Tcl_Interp *interp, int objc, 
	       struct Tcl_Obj * CONST objv[])
{
    BYTE key;
    int on,state;
    char *arg;

    if (objc > 1) {
	arg = Tcl_GetString(objv[1]);
	switch (*arg) {
	case 'c':
	    key = VK_CAPITAL;
	    break;
	case 'n':
	    key = VK_NUMLOCK;
	    break;
	case 's':
	    key = VK_SCROLL;
	    break;
	default:
	    goto Usage;
	}
    } else {
Usage:
	Tcl_WrongNumArgs(interp, 1, objv, "caps|num|scroll [boolean]");
	return TCL_ERROR;
    }

    state = GetLock(key);

    if (objc == 3) {
	if (Tcl_GetBooleanFromObj(interp, objv[2], &on) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (on ^ state) {
	    TogLock(key);
	    state = GetLock(key);
	}
    }

    Tcl_SetObjResult(interp, Tcl_NewBooleanObj(state));
    return TCL_OK;
}

int
Keylock_Init(Tcl_Interp *interp)
{
    Tcl_CreateObjCommand(interp, "winutils::keylock", KeyLockObjCmd, 0L, 0L);
    return TCL_OK;
}