/* ----------------------------------------------------------------------------
 * Launch.c --
 *
 *	The launch and exetype commands.  Fully thread safe.
 *
 * ----------------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this.  See license.txt
 * for more detail.
 *
 * ----------------------------------------------------------------------------
 * RCS: @(#) $Id: Launch.c,v 1.25 2003/04/06 23:27:49 davygrvy Exp $
 * ----------------------------------------------------------------------------
 */

#include "tcl.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

/* Platform SDK specifics we can avoid */
#ifndef INVALID_SET_FILE_POINTER
#   define INVALID_SET_FILE_POINTER ((DWORD)-1)
#endif
#ifndef BELOW_NORMAL_PRIORITY_CLASS
#   define BELOW_NORMAL_PRIORITY_CLASS       0x00004000
#endif
#ifndef ABOVE_NORMAL_PRIORITY_CLASS
#   define ABOVE_NORMAL_PRIORITY_CLASS       0x00008000
#endif

typedef struct {
    int useWide;
    HANDLE (WINAPI *createFileProc)(const TCHAR *, DWORD, DWORD, 
	    LPSECURITY_ATTRIBUTES, DWORD, DWORD, HANDLE);
    BOOL (WINAPI *createProcessProc)(const TCHAR *, TCHAR *, 
	    LPSECURITY_ATTRIBUTES, LPSECURITY_ATTRIBUTES, BOOL, DWORD, 
	    LPVOID, const TCHAR *, LPSTARTUPINFO, LPPROCESS_INFORMATION);
    DWORD (WINAPI *getFileAttributesProc)(const TCHAR *);
    DWORD (WINAPI *getShortPathNameProc)(const TCHAR *, TCHAR *, DWORD); 
    DWORD (WINAPI *searchPathProc)(const TCHAR *, const TCHAR *, 
	    const TCHAR *, DWORD, TCHAR *, TCHAR **);
} WinProcs;

static WinProcs asciiProcs = {
    0,
    CreateFileA,
    CreateProcessA,
    GetFileAttributesA,
    GetShortPathNameA,
    SearchPathA
};

static WinProcs unicodeProcs = {
    1,
    (void *) CreateFileW,
    (void *) CreateProcessW,
    (void *) GetFileAttributesW,
    (void *) GetShortPathNameW,
    (void *) SearchPathW
};

WinProcs *winProcs = &asciiProcs;

TCL_DECLARE_MUTEX(initLock)
static int initialized = 0;

/*
 *  All executable types.
 */

#define APPL_NONE	    0
#define APPL_BATCH	    1
#define APPL_DOS16	    2
#define APPL_OS2	    3
#define APPL_OS2DRV	    4
#define APPL_WIN16	    5
#define APPL_WIN16DRV	    6
#define APPL_WIN32CUI	    7
#define APPL_WIN32GUI	    8
#define APPL_WIN32DLL	    9
#define APPL_WIN32DRV	    10
#define APPL_WIN64CUI	    11
#define APPL_WIN64GUI	    12
#define APPL_WIN64DLL	    13
#define APPL_WIN64DRV	    14


/* Local Protos */
static Tcl_Obj *GetSysMsgObj(DWORD id);
static int WinApplicationType(const char *originalName, Tcl_DString *fullName);
static void BuildCommandLine(const char *executable, int objc,
	struct Tcl_Obj * const objv[], Tcl_DString *linePtr);

Tcl_Obj *GetSysMsgObj(DWORD id)
{
    int chars;
    char sysMsgSpace[512];

    chars = wsprintf(sysMsgSpace, "[%u] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_IGNORE_INSERTS |
	    FORMAT_MESSAGE_MAX_WIDTH_MASK, 0L, id, 0, sysMsgSpace+chars, (512-chars),
	    0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}


int
LaunchObjCmd (ClientData cData, Tcl_Interp *interp, int objc,
	     struct Tcl_Obj * const objv[])
{
    static char *options[] = {"-desktop", "-ctitle", "-cdims", "-cfill",
	    "-cfull", "-chide", "-origin", "-size", "-show", "-priority",
	    "-wait", "-directory", "-env", 0L};
    enum options {LAUNCH_DESK, LAUNCH_TITLE, LAUNCH_CDIMS, LAUNCH_CFILL,
	    LAUNCH_CFULL, LAUNCH_CHIDE, LAUNCH_ORIGIN, LAUNCH_SIZE, LAUNCH_SHOW,
	    LAUNCH_PRIORITY, LAUNCH_WAIT, LAUNCH_DIR, LAUNCH_ENV};
    int option;
    static char *sstyles[] = { "hide", "maximum", "minimum", "normal", NULL };
    enum sstyles { SHOW_HIDE, SHOW_MAX, SHOW_MIN, SHOW_NORM };
    int sstyle;
    static char *priorities[] = { "realtime", "high", "above", "normal", "below",
	    "idle", NULL };
    enum priorities { PRIORITY_REAL, PRIORITY_HIGH, PRIORITY_ABOVE, PRIORITY_NORM,
	    PRIORITY_BELOW, PRIORITY_IDLE };
    int priority;
    int j, len, result = TCL_ERROR, done = 0, applType;
    const char *arg;
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    Tcl_DString cmdLine, exeFullPathUTF, deskTop, dir, title, env;
    DWORD dwCreationFlags, doWait = FALSE;
    Tcl_Obj *file = 0L, *directory = 0L;
    char *p;
    TCHAR *lpDirectory = 0L;
    LPVOID envBlock;

    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
    ZeroMemory(&si, sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);

    Tcl_DStringInit(&deskTop);
    Tcl_DStringInit(&title);
    Tcl_DStringInit(&dir);
    Tcl_DStringInit(&cmdLine);
    Tcl_DStringInit(&exeFullPathUTF);
    Tcl_DStringInit(&env);

    dwCreationFlags = CREATE_DEFAULT_ERROR_MODE | NORMAL_PRIORITY_CLASS;

    /* Decode args */
    if (objc > 1) {
	for (j = 1; (j < objc) && !done; j++) {
	    arg = Tcl_GetString(objv[j]);
	    if (*arg == '-') {
		if (Tcl_GetIndexFromObj(interp, objv[j], options, "option", 0,
			&option) != TCL_OK) {
		    goto end;
		}
		switch ((enum options) option) {
		case LAUNCH_DESK:
		    if (objc > j+1) {
			p = Tcl_GetStringFromObj(objv[++j], &len);
			si.lpDesktop = (LPSTR) Tcl_WinUtfToTChar(p, len, &deskTop);
			break;
		    } else {
			Tcl_SetObjResult(interp, Tcl_NewStringObj(
				"The -desktop option requires a parameter.", -1));
			goto end;
		    }
		case LAUNCH_TITLE:
		    if (objc > j+1) {
			p = Tcl_GetStringFromObj(objv[++j], &len);
			si.lpTitle = (LPSTR) Tcl_WinUtfToTChar(p, len, &title);
			break;
		    } else {
			Tcl_SetObjResult(interp, Tcl_NewStringObj(
				"The -ctitle option requires a parameter.", -1));
			goto end;
		    }
		case LAUNCH_CDIMS:
		    if (objc > j+2) {
			if (Tcl_GetLongFromObj(interp, objv[++j], &si.dwXCountChars)
				== TCL_ERROR) {
			    goto end;
			}
			if (Tcl_GetLongFromObj(interp, objv[++j], &si.dwYCountChars)
				== TCL_ERROR) {
			    goto end;
			}
			si.dwFlags |= STARTF_USECOUNTCHARS;
			break;
		    } else {
			Tcl_SetObjResult(interp, Tcl_NewStringObj(
				"The -cdims option requires 2 parameters.", -1));
			goto end;
		    }
		case LAUNCH_CFILL:
		    if (objc > j+1) {
			if (Tcl_GetLongFromObj(interp, objv[++j], &si.dwFillAttribute)
				== TCL_ERROR) {
			    goto end;
			}
			if (si.dwFillAttribute < 0 || si.dwFillAttribute > 255) {
			    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				    Tcl_GetString(objv[j]), " is out-of-range (0-255)", 0L);
			    goto end;
			}
			si.dwFlags |= STARTF_USEFILLATTRIBUTE;
			break;
		    } else {
			Tcl_SetObjResult(interp, Tcl_NewStringObj(
				"The -cfill option requires a parameter.", -1));
			goto end;
		    }
		case LAUNCH_CFULL:
		    si.dwFlags |= STARTF_RUNFULLSCREEN;
		    break;
		case LAUNCH_CHIDE:
		    dwCreationFlags |= CREATE_NO_WINDOW;
		    break;
		case LAUNCH_ORIGIN:
		    if (objc > j+2) {
			if (Tcl_GetLongFromObj(interp, objv[++j], &si.dwX)
				== TCL_ERROR) {
			    goto end;
			}
			if (Tcl_GetLongFromObj(interp, objv[++j], &si.dwY)
				== TCL_ERROR) {
			    goto end;
			}
			si.dwFlags |= STARTF_USEPOSITION;
			break;
		    } else {
			Tcl_SetObjResult(interp, Tcl_NewStringObj(
				"The -origin option requires 2 parameters.", -1));
			goto end;
		    }
		case LAUNCH_SIZE:
		    if (objc > j+2) {
			if (Tcl_GetLongFromObj(interp, objv[++j], &si.dwXSize)
				== TCL_ERROR) {
			    goto end;
			}
			if (Tcl_GetLongFromObj(interp, objv[++j], &si.dwYSize)
				== TCL_ERROR) {
			    goto end;
			}
			si.dwFlags |= STARTF_USESIZE;
			break;
		    } else {
			Tcl_SetObjResult(interp, Tcl_NewStringObj(
				"The -size option requires 2 parameters.", -1));
			goto end;
		    }
		case LAUNCH_SHOW: /* show style */
		    if (objc>j+1) {
			if (Tcl_GetIndexFromObj(interp, objv[++j], sstyles,
				"ShowStyle", 0, &sstyle) != TCL_OK) {
			    goto end;
			}

			switch ((enum sstyles) sstyle) {
			case SHOW_HIDE:
			    si.wShowWindow = SW_HIDE; break;
			case SHOW_MAX:
			    si.wShowWindow = SW_SHOWMAXIMIZED; break;
			case SHOW_MIN:
			    si.wShowWindow = SW_SHOWMINNOACTIVE; break;
			case SHOW_NORM:
			    si.wShowWindow = SW_SHOWNORMAL;
			}
			si.dwFlags |= STARTF_USESHOWWINDOW;
			break;
		    } else {
			Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				"The -show option requires a parameter.  ", 0L);
			goto end;
		    }
		case LAUNCH_PRIORITY: /* priority */
		    if (objc>j+1) {
			if (Tcl_GetIndexFromObj(interp, objv[++j], priorities,
				"PriorityClass", 0, &priority) != TCL_OK) {
			    goto end;
			}

			/* suck-out the normal first */
			dwCreationFlags &= ~(NORMAL_PRIORITY_CLASS);

			switch ((enum priorities) priority) {
			case PRIORITY_REAL:
			    dwCreationFlags |= REALTIME_PRIORITY_CLASS; break;
			case PRIORITY_HIGH:
			    dwCreationFlags |= HIGH_PRIORITY_CLASS; break;
			case PRIORITY_ABOVE:
			    dwCreationFlags |= ABOVE_NORMAL_PRIORITY_CLASS; break;
			case PRIORITY_NORM:
			    dwCreationFlags |= NORMAL_PRIORITY_CLASS; break;
			case PRIORITY_BELOW:
			    dwCreationFlags |= BELOW_NORMAL_PRIORITY_CLASS; break;
			case PRIORITY_IDLE:
			    dwCreationFlags |= IDLE_PRIORITY_CLASS;
			}
			break;
		    } else {
			Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				"The -priority option requires a parameter.  ", 0L);
			goto end;
		    }
		case LAUNCH_WAIT: /* wait */
		    doWait = TRUE;
		    break;

		case LAUNCH_DIR:
		    if (objc>j+1) {
			directory = Tcl_DuplicateObj(objv[++j]);
			for (p = Tcl_GetString(directory); *p != '\0'; p++) {
			    if (*p == '/') *p = '\\';
			}
			p = Tcl_GetStringFromObj(directory, &len);
			lpDirectory = (LPSTR) Tcl_WinUtfToTChar(p, len, &dir);
			Tcl_DecrRefCount(directory);
		    } else {
			Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				"The -directory option requires a parameter.  ", 0L);
			goto end;
		    }
		    break;

		case LAUNCH_ENV: /* env */
		    if (objc>j+1) {

			Tcl_CmdInfo arrayCmdInfo;
			Tcl_Obj *pairsListObj, *cmd[3], **elemArray;
			int i, listLen;
			Tcl_DString temp;

			/*
			 *  We need to call the array command so traces get run
			 *  correctly.  No direct access seems to be available
			 *  anyways unless I go right to the hash table.
			 */

			Tcl_GetCommandInfo(interp, "array", &arrayCmdInfo);
			cmd[0] = Tcl_NewStringObj("array", -1);
			cmd[1] = Tcl_NewStringObj("get", -1);
			cmd[2] = objv[++j];
			if ((arrayCmdInfo.objProc)(arrayCmdInfo.objClientData,
				interp, 3, cmd)	!= TCL_OK)
			{
			    goto end;
			}
			pairsListObj = Tcl_GetObjResult(interp);
			Tcl_IncrRefCount(pairsListObj);
			Tcl_ResetResult(interp);
			Tcl_DecrRefCount(cmd[0]);
			Tcl_DecrRefCount(cmd[1]);

			if (Tcl_ListObjGetElements(0L, pairsListObj, &listLen,
				&elemArray) != TCL_OK || !listLen)
			{
			    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				    "The array variable named \"",
				    Tcl_GetString(cmd[2]),
				    "\" was either empty or not an array.  ",
				    0L);
			    Tcl_DecrRefCount(pairsListObj);
			    goto end;
			}

			/*
			 *  Format is a null seperated, double null terminated
			 *  list. ex. "key=val\0key=val\0"
			 */

			Tcl_DStringInit(&temp);
			for (i=0; i<listLen; i+=2) {
			    int len;
			    char *data;

			    data = Tcl_GetStringFromObj(elemArray[i], &len);
			    Tcl_WinUtfToTChar(data, len, &temp);
			    Tcl_DStringAppend(&env, Tcl_DStringValue(&temp),
				    Tcl_DStringLength(&temp));

			    /* Convert it to unicode the easy way */
			    Tcl_DStringAppend(&env, "=",
				    (winProcs->useWide?2:1));

			    data = Tcl_GetStringFromObj(elemArray[i+1], &len);
			    Tcl_WinUtfToTChar(data, len, &temp);
			    Tcl_DStringAppend(&env, Tcl_DStringValue(&temp),
				    Tcl_DStringLength(&temp));

			    Tcl_DStringAppend(&env, "\0",
				    (winProcs->useWide?2:1));
			}
			Tcl_DStringFree(&temp);
			Tcl_DecrRefCount(pairsListObj);

			if (winProcs->useWide)
				Tcl_DStringAppend(&env, "\0", 1);

		    } else {

			Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				"The -env option requires a parameter.  ", 0L);
			goto end;
		    }
		    break;
		}
	    } else {
		done = 1;
		/* back it up one. */
		--j;
	    }
	}

	if (objc == j) {
	    Tcl_SetObjResult(interp, Tcl_NewStringObj("Nothing to execute.",
		    -1));
	    goto end;
	}

	file = Tcl_DuplicateObj(objv[j++]);
	for (p = Tcl_GetString(file); *p != '\0'; p++) {
	    if (*p == '/') *p = '\\';
	}

    } else {
	Tcl_WrongNumArgs(interp, 1, objv, "[options] <exe> [<params> ...]");
	goto end;
    }

    applType = WinApplicationType(Tcl_GetString(file), &exeFullPathUTF);
    Tcl_DecrRefCount(file);

    switch (applType) {
    case APPL_NONE:
       	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	goto end;

    case APPL_BATCH:
       	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"Sorry, can't run batchfiles, yet.", 0L);
	goto end;

    case APPL_WIN16:
    case APPL_WIN32GUI:
    case APPL_WIN64GUI:
	if (si.dwFlags & STARTF_USEFILLATTRIBUTE ||
	    si.dwFlags & STARTF_RUNFULLSCREEN ||
	    si.dwFlags & STARTF_USECOUNTCHARS)
	{
       	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		    "\"", Tcl_DStringValue(&exeFullPathUTF),
		    "\" is a GUI app and can't take console attributes.", 0L);
	    goto end;
	}
	break;

    case APPL_DOS16:
       	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"Sorry, can't run DOS-16 apps, yet.", 0L);
	goto end;

    case APPL_OS2:
    case APPL_WIN32CUI:
    case APPL_WIN64CUI:
	dwCreationFlags |= CREATE_NEW_CONSOLE;
	break;
    }

    /*
     *  Properly quote params into a single string and translate
     *  that string into the proper TCHAR rep.
     */
    BuildCommandLine(Tcl_DStringValue(&exeFullPathUTF), objc-j, &objv[j],
	    &cmdLine);

    /*
     *  If a custom environment block isn't being used, set the block pointer
     *  to NULL to indicate to CreateProcess() that we want the current
     *  parent's block to be inherited.  As opposed to a pointer to an empty
     *  string which I think means a blank environment.
     */
    if (*(Tcl_DStringValue(&env))) {
	envBlock = Tcl_DStringValue(&env);
	if (winProcs->useWide)
		dwCreationFlags |= CREATE_UNICODE_ENVIRONMENT;
    } else {
	envBlock = 0L;
    }

    /* Launch it */
    if (!winProcs->createProcessProc(
	0L,		    /* No module name used. */
	(TCHAR *) Tcl_DStringValue(&cmdLine), /* Command line. */
	0L,		    /* Process handle not inheritable. */
	0L,		    /* Thread handle not inheritable. */
	FALSE,		    /* Set handle inheritance to FALSE. */
	dwCreationFlags,    /* Creation flags. */
	envBlock,	    /* Use specified environment block. */
	lpDirectory,	    /* Starting directory. */
	&si,		    /* Pointer to STARTUPINFO structure. */
	&pi))		    /* Pointer to PROCESS_INFORMATION structure. */
    {
       	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	goto end;
    }

    /* This wait is used to force the OS to give some time to the character-mode
     * process. */

    if (applType == APPL_DOS16) {
	WaitForSingleObject(pi.hProcess, 50);
    }

    /* 
     * "When an application spawns a process repeatedly, a new thread 
     * instance will be created for each process but the previous 
     * instances may not be cleaned up.  This results in a significant 
     * virtual memory loss each time the process is spawned.  If there 
     * is a WaitForInputIdle() call between CreateProcess() and
     * CloseHandle(), the problem does not occur." PSS ID Number: Q124121
     */

    WaitForInputIdle(pi.hProcess, 5000);
    CloseHandle(pi.hThread);

    if (doWait) {
	WaitForSingleObject(pi.hProcess, INFINITE);
	GetExitCodeProcess(pi.hProcess, &doWait);
	Tcl_SetObjResult(interp, Tcl_NewLongObj(doWait));
    } else {
	Tcl_SetObjResult(interp, Tcl_NewLongObj(pi.dwProcessId));
    }

    /* Remove the reference we have been given. */
    CloseHandle(pi.hProcess);

    result = TCL_OK;

end:
    Tcl_DStringFree(&deskTop);
    Tcl_DStringFree(&title);
    Tcl_DStringFree(&dir);
    Tcl_DStringFree(&cmdLine);
    Tcl_DStringFree(&exeFullPathUTF);
    Tcl_DStringFree(&env);
    return result;
}

int
ExeTypeObjCmd (ClientData cData, Tcl_Interp *interp, int objc,
	       struct Tcl_Obj * const objv[])
{
    Tcl_DString fullName;
    char *p;
    const char *type;
    Tcl_Obj *data[2], *error = 0L;

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "<exe>");
	return TCL_ERROR;
    }

    Tcl_DStringInit(&fullName);

    switch (WinApplicationType(Tcl_GetString(objv[1]), &fullName)) {
	case APPL_NONE:
	    type = "None";
       	    error = GetSysMsgObj(GetLastError());
	    break;
	case APPL_BATCH:
	    type = "Batch"; break;
	case APPL_DOS16:
	    type = "MS-DOS"; break;
	case APPL_WIN16:
	    type = "Win16 GUI"; break;
	case APPL_WIN16DRV:
	    type = "Win16 Driver"; break;
	case APPL_OS2:
	    type = "OS/2"; break;
	case APPL_OS2DRV:
	    type = "OS/2 Driver"; break;
	case APPL_WIN32CUI:
	    type = "Win32 CUI"; break;
	case APPL_WIN32GUI:
	    type = "Win32 GUI"; break;
	case APPL_WIN32DLL:
	    type = "Win32 DLL"; break;
	case APPL_WIN32DRV:
	    type = "Win32 Driver"; break;
	case APPL_WIN64CUI:
	    type = "Win64 CUI"; break;
	case APPL_WIN64GUI:
	    type = "Win64 GUI"; break;
	case APPL_WIN64DLL:
	    type = "Win64 DLL"; break;
	case APPL_WIN64DRV:
	    type = "Win64 Driver"; break;
    }

    for (p = Tcl_DStringValue(&fullName); *p != '\0'; p++) {
	if (*p == '\\') {
	    *p = '/';
	}
    }

    data[0] = Tcl_NewStringObj(type, -1);
    data[1] = Tcl_NewStringObj(Tcl_DStringValue(&fullName), -1);
    Tcl_DStringFree(&fullName);
    Tcl_SetObjResult(interp, Tcl_NewListObj(2, data));
    if (error) {
	Tcl_ListObjAppendElement(0L, Tcl_GetObjResult(interp), error); 
    }
    return TCL_OK;
}

static int
WinApplicationType(
    const char *originalName,	/* Name of the application to find. (in UTF-8) */
    Tcl_DString *fullName)	/* Buffer space filled with complete path to 
				 * application. (in UTF-8) */
{
    int applType, i, nameLen, nativeNameLen;
    HANDLE hFile;
    TCHAR *rest;
    char *ext;
    DWORD attr, read;
    IMAGE_DOS_HEADER p236;  /* p236, DOS (old-style) executable-file header */
    union {
	BYTE buf[200];
	IMAGE_NT_HEADERS pe;
	IMAGE_OS2_HEADER ne;
	IMAGE_VXD_HEADER le;
    } header;
    Tcl_DString nameBuf, nativeNameBuff, ds;
    const TCHAR *nativeName;
    WCHAR nativeShortPath[MAX_PATH];   /* needed for unicode space */
    static char extensions[][5] = {"", ".com", ".exe", ".bat", ".cmd"};
    int offset64;

    /* Look for the program as an external program.  First try the name
     * as it is, then try adding .com, .exe, and .bat, in that order, to
     * the name, looking for an executable.
     *
     * Using the raw SearchPath() procedure doesn't do quite what is 
     * necessary.  If the name of the executable already contains a '.' 
     * character, it will not try appending the specified extension when
     * searching (in other words, SearchPath will not find the program 
     * "a.b.exe" if the arguments specified "a.b" and ".exe").   
     * So, first look for the file as it is named.  Then manually append 
     * the extensions, looking for a match.  
     */

    applType = APPL_NONE;
    Tcl_DStringInit(&nameBuf);
    Tcl_DStringInit(&nativeNameBuff);
    Tcl_DStringAppend(&nameBuf, originalName, -1);
    Tcl_DStringAppend(fullName, originalName, -1);
    nameLen = Tcl_DStringLength(&nameBuf);

    for (i = 0; i < (sizeof(extensions) / sizeof(extensions[0])); i++) {
	Tcl_DStringSetLength(&nameBuf, nameLen);
	Tcl_DStringAppend(&nameBuf, extensions[i], -1);
        nativeName = Tcl_WinUtfToTChar(Tcl_DStringValue(&nameBuf), 
		Tcl_DStringLength(&nameBuf), &ds);

	/* Just get the size of the buffer needed, when found. */
	nativeNameLen = winProcs->searchPathProc(NULL, nativeName,
		NULL, 0, NULL, &rest);

	if (nativeNameLen == 0) {
	    /* not found. */
	    Tcl_DStringFree(&ds);
	    continue;
	}

	/* Set the buffer needed. */
	Tcl_DStringSetLength(&nativeNameBuff, 
		(winProcs->useWide ? nativeNameLen*2 : nativeNameLen));

	winProcs->searchPathProc(NULL, nativeName, NULL,
		Tcl_DStringLength(&nativeNameBuff),
		(TCHAR *) Tcl_DStringValue(&nativeNameBuff), &rest);
	Tcl_DStringFree(&ds);

	/*
	 * Ignore matches on directories, keep falling through
	 * when identified as something else.
	 */

	attr = winProcs->getFileAttributesProc(
		(TCHAR *) Tcl_DStringValue(&nativeNameBuff));
	if ((attr == -1) || (attr & FILE_ATTRIBUTE_DIRECTORY)) {
	    continue;
	}
	Tcl_WinTCharToUtf((TCHAR *) Tcl_DStringValue(&nativeNameBuff),
		-1, fullName);

	ext = strrchr(Tcl_DStringValue(fullName), '.');
	if ((ext != NULL) && (stricmp(ext, ".bat") == 0 ||
		stricmp(ext, ".cmd") == 0)) {
	    applType = APPL_BATCH;
	    break;
	}
	
	hFile = winProcs->createFileProc(
		(TCHAR *) Tcl_DStringValue(&nativeNameBuff),
		GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 
		FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
	    applType = APPL_NONE;
	    break;
	}

	p236.e_magic = 0;
	ReadFile(hFile, &p236, sizeof(IMAGE_DOS_HEADER), &read, NULL);
	if (p236.e_magic != IMAGE_DOS_SIGNATURE) {
	    /* 
	     * Doesn't have the magic number for relocatable executables.  If 
	     * filename ends with .com, assume it's a DOS application anyhow.
	     * Note that we didn't make this assumption at first, because some
	     * supposed .com files are really 32-bit executables with all the
	     * magic numbers and everything.  
	     */

	    /*
	     * Additional notes from Ralf Brown's interupt list:
	     *
	     * The COM files are raw binary executables and are a leftover
	     * from the old CP/M machines with 64K RAM.  A COM program can
	     * only have a size of less than one segment (64K), including code
	     * and static data since no fixups for segment relocation or
	     * anything else is included. One method to check for a COM file
	     * is to check if the first byte in the file could be a valid jump
	     * or call opcode, but this is a very weak test since a COM file
	     * is not required to start with a jump or a call. In principle,
	     * a COM file is just loaded at offset 100h in the segment and
	     * then executed.
	     *
	     * OFFSET              Count TYPE   Description
	     * 0000h                   1 byte   ID=0E9h
	     *                                  ID=0EBh
	     */

	    CloseHandle(hFile);
	    if ((ext != NULL) && (stricmp(ext, ".com") == 0)) {
		applType = APPL_DOS16;
		break;
	    }
	    SetLastError(ERROR_INVALID_EXE_SIGNATURE);
	    applType = APPL_NONE;
	    break;
	}
	if (p236.e_lfarlc < 0x40 || p236.e_lfanew == 0 /* reserved */) {
	    /* 
	     * Old-style header only.  Can't be more than a DOS16 executable.
	     */

	    CloseHandle(hFile);
	    applType = APPL_DOS16;
	    break;
	}

	/* 
	 * The LONG at p236.e_lfanew points to the real exe header only
	 * when p236.e_lfarlc is set to 40h (or greater).
	 */
	
	if (SetFilePointer(hFile, p236.e_lfanew, NULL, FILE_BEGIN)
		== INVALID_SET_FILE_POINTER) {
	    /* Bogus PE header pointer. */
	    CloseHandle(hFile);
	    SetLastError(ERROR_BAD_EXE_FORMAT);
	    applType = APPL_NONE;
	    break;
	}
	ReadFile(hFile, header.buf, 200, &read, NULL);
	CloseHandle(hFile);

	/*
	 * Check the sigs against the following list:
	 *  'PE\0\0'  Win32 (Windows NT and Win32s) portable executable based
	 *	    on Unix COFF.
	 *  'NE'  Windows or OS/2 1.x segmented ("new") executable.
	 *  'LE'  Windows virtual device driver (VxD) linear executable.
	 *  'LX'  Variant of LE used in OS/2 2.x
	 *  'W3'  Windows WIN386.EXE file; a collection of LE files
	 *	    (protected mode windows).
	 *  'W4'  Variant of above.
	 */

	if (header.pe.Signature == IMAGE_NT_SIGNATURE) {
	    if (!(header.pe.FileHeader.Characteristics &
		    IMAGE_FILE_EXECUTABLE_IMAGE)) {
		/* Not an executable. */
		SetLastError(ERROR_BAD_EXE_FORMAT);
		applType = APPL_NONE;
		break;
	    }

	    if (header.pe.OptionalHeader.Magic ==
		    IMAGE_NT_OPTIONAL_HDR32_MAGIC) {
		/* Win32 executable */
		offset64 = 0;
	    } else if (header.pe.OptionalHeader.Magic ==
		    IMAGE_NT_OPTIONAL_HDR64_MAGIC) {
		/* Win64 executable */
		offset64 = 4;
	    } else {
		/* Unknown magic number */
		SetLastError(ERROR_INVALID_MODULETYPE);
		applType = APPL_NONE;
		break;
	    }

	    if (header.pe.FileHeader.Characteristics & IMAGE_FILE_DLL) {
		/*
		 * DLLs are executable, but indirectly.  We shouldn't return
		 * APPL_NONE or the subsystem that its said to run under
		 * as it's not the complete truth, so return a new type and
		 * let the user decide what to do.
		 */

		applType = APPL_WIN32DLL + offset64;
		break;
	    }

	    switch (header.pe.OptionalHeader.Subsystem) {
		case IMAGE_SUBSYSTEM_WINDOWS_CUI:
		case IMAGE_SUBSYSTEM_OS2_CUI:
		case IMAGE_SUBSYSTEM_POSIX_CUI:
		    /* Runs in the CUI subsystem */
		    applType = APPL_WIN32CUI + offset64;
		    break;

		case IMAGE_SUBSYSTEM_WINDOWS_GUI:
		case IMAGE_SUBSYSTEM_WINDOWS_CE_GUI:
		    /* Runs in the GUI subsystem */
		    applType = APPL_WIN32GUI + offset64;
		    break;

		case IMAGE_SUBSYSTEM_UNKNOWN:
		case IMAGE_SUBSYSTEM_NATIVE:
		case IMAGE_SUBSYSTEM_NATIVE_WINDOWS:
		    /* Special Driver */
		    applType = APPL_WIN32DRV + offset64;
		    break;
	    }

#define IMAGE_NE_FLAG_DRIVER	0x8000
#define IMAGE_NE_EXETYP_OS2	0x1
#define IMAGE_NE_EXETYP_WIN	0x2
#define IMAGE_NE_EXETYP_DOS4X	0x3
#define IMAGE_NE_EXETYP_WIN386	0x4

	} else if (header.ne.ne_magic == IMAGE_OS2_SIGNATURE) {
	    switch (header.ne.ne_exetyp) {
		case IMAGE_NE_EXETYP_OS2:    /* Microsoft/IBM OS/2 1.x */
		    if (header.ne.ne_flags & IMAGE_NE_FLAG_DRIVER) {
			applType = APPL_OS2DRV;
		    } else {
			applType = APPL_OS2;
		    }
		    break;

		case IMAGE_NE_EXETYP_WIN:    /* Microsoft Windows */
		case IMAGE_NE_EXETYP_WIN386: /* Same, but Protected mode */
		    if (header.ne.ne_flags & IMAGE_NE_FLAG_DRIVER) {
			applType = APPL_WIN16DRV;
		    } else {
			applType = APPL_WIN16;
		    }
		    break;

		case IMAGE_NE_EXETYP_DOS4X:  /* Microsoft MS-DOS 4.x */
		    applType = APPL_DOS16;
		    break;

		default:
		    /* Unidentified */
		    SetLastError(ERROR_INVALID_MODULETYPE);
		    applType = APPL_NONE;
	    }
	} else if (
		header.le.e32_magic == IMAGE_OS2_SIGNATURE_LE /* 'LE' */ ||
		header.le.e32_magic == 0x584C /* 'LX' */ ||
		header.le.e32_magic == 0x3357 /* 'W3' */ ||
		header.le.e32_magic == 0x3457 /* 'W4' */
	){
	    /* Virtual device drivers are not executables, per se. */
	    applType = APPL_WIN16DRV;
	} else {
	    /* The loader will barf anyway, so barf now. */
	    SetLastError(ERROR_INVALID_EXE_SIGNATURE);
	    applType = APPL_NONE;
	}
	break;
    }

    if (applType == APPL_DOS16 || applType == APPL_WIN16 ||
	    applType == APPL_WIN16DRV || applType == APPL_OS2 ||
	    applType == APPL_OS2DRV) {
	/* 
	 * Replace long path name of executable with short path name for
	 * 16-bit applications.  Otherwise the application may not be able
	 * to correctly parse its own command line to separate off the
	 * application name from the arguments.
	 */

	winProcs->getShortPathNameProc(
		(TCHAR *) Tcl_DStringValue(&nativeNameBuff),
		(TCHAR *) nativeShortPath, MAX_PATH);
	Tcl_WinTCharToUtf((TCHAR *) nativeShortPath, -1, fullName);
    }
    Tcl_DStringFree(&nativeNameBuff);
    Tcl_DStringFree(&nameBuf);
    return applType;
}

/*    
 *----------------------------------------------------------------------
 *
 * BuildCommandLine --
 *
 *	The command line arguments are stored in linePtr separated
 *	by spaces, in a form that CreateProcess() understands.  Special 
 *	characters in individual arguments from argv[] must be quoted 
 *	when being stored in cmdLine.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 * Comment: COPY OF NON_PUBLIC CORE FUNCTION WITH CHANGES!
 *
 *----------------------------------------------------------------------
 */
static void
BuildCommandLine(
    const char *executable,	/* Full path of executable (including 
				 * extension).  Replacement for argv[0]. */
    int objc,			/* Number of arguments. */
    struct Tcl_Obj * const objv[], /* Array of args. */
    Tcl_DString *linePtr)	/* Initialized Tcl_DString that receives the
				 * command line (TCHAR). */
{
    const char *arg, *start, *special;
    int quote, i;
    Tcl_DString ds;

    Tcl_DStringInit(&ds);

    /*
     * Prime the path.
     */
    
    Tcl_DStringAppend(&ds, Tcl_DStringValue(linePtr), -1);
    
    for (i = -1; i < objc; i++) {
	if (i == -1) {
	    arg = executable;
	} else {
	    arg = Tcl_GetString(objv[i]);
	    Tcl_DStringAppend(&ds, " ", 1);
	}

	quote = 0;
	if (arg[0] == '\0') {
	    quote = 1;
	} else {
	    for (start = arg; *start != '\0'; start++) {
		if (isspace(*start)) { /* INTL: ISO space. */
		    quote = 1;
		    break;
		}
	    }
	}
	if (quote) {
	    Tcl_DStringAppend(&ds, "\"", 1);
	}

	start = arg;	    
	for (special = arg; ; ) {
	    if ((*special == '\\') && 
		    (special[1] == '\\' || special[1] == '"')) {
		Tcl_DStringAppend(&ds, start, (int) (special - start));
		start = special;
		while (1) {
		    special++;
		    if (*special == '"') {
			/* 
			 * N backslashes followed a quote -> insert 
			 * N * 2 + 1 backslashes then a quote.
			 */

			Tcl_DStringAppend(&ds, start, (int) (special - start));
			break;
		    }
		    if (*special != '\\') {
			break;
		    }
		}
		Tcl_DStringAppend(&ds, start, (int) (special - start));
		start = special;
	    }
	    if (*special == '"') {
		Tcl_DStringAppend(&ds, start, (int) (special - start));
		Tcl_DStringAppend(&ds, "\\\"", 2);
		start = special + 1;
	    }
	    if (*special == '\0') {
		break;
	    }
	    special++;
	}
	Tcl_DStringAppend(&ds, start, (int) (special - start));
	if (quote) {
	    Tcl_DStringAppend(&ds, "\"", 1);
	}
    }
    Tcl_DStringFree(linePtr);
    Tcl_WinUtfToTChar(Tcl_DStringValue(&ds), Tcl_DStringLength(&ds), linePtr);
    Tcl_DStringFree(&ds);
}


int
Launch_Init(Tcl_Interp *interp)
{
    Tcl_MutexLock(&initLock);

    if (!initialized) {
	OSVERSIONINFO os;

	initialized = 1;
	os.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&os);
	if (os.dwPlatformId == VER_PLATFORM_WIN32_NT) {    
	    winProcs = &unicodeProcs;
	}
    }

    Tcl_MutexUnlock(&initLock);
    Tcl_CreateObjCommand(interp, "winutils::launch", LaunchObjCmd, 0L, 0L);
    Tcl_CreateObjCommand(interp, "winutils::exetype", ExeTypeObjCmd, 0L, 0L);
    return TCL_OK;
}
