/* ----------------------------------------------------------------------
 * PerfMon.c --
 *
 *	Get Performance monitor counter by name.
 *
 * ----------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this code.  See
 * license.txt for more detail.
 *
 * ----------------------------------------------------------------------
 * RCS: @(#) $Id: PerfMon.c,v 1.3 2002/05/22 00:14:37 davygrvy Exp $
 * ----------------------------------------------------------------------
 */

#include "tcl.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <pdh.h>

TCL_DECLARE_MUTEX(initLock)
static int initialized = 0;
static int usePerfMon = 0;
static int gotWideInt = 0;
static HMODULE hPDH;

struct {
    PDH_STATUS (__stdcall *PdhOpenQueryProc)
	    (LPCSTR, DWORD_PTR, PDH_HQUERY *);
    PDH_STATUS (__stdcall *PdhAddCounterProc)
	    (PDH_HQUERY, LPCSTR, DWORD_PTR, PDH_HCOUNTER *);
    PDH_STATUS (__stdcall *PdhCollectQueryDataProc)
	    (PDH_HQUERY);
    PDH_STATUS (__stdcall *PdhGetFormattedCounterValueProc)
	    (PDH_HCOUNTER, DWORD, LPDWORD, PPDH_FMT_COUNTERVALUE);
    PDH_STATUS (__stdcall *PdhCloseQueryProc)
	    (PDH_HQUERY);
} pdhProcs;


static Tcl_Obj *
GetPdhMsgObj(PDH_STATUS id)
{
    int chars;
    char sysMsgSpace[512];  /* <- stack space abuse! */

    chars = wsprintf(sysMsgSpace, "[#%x] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_HMODULE|
	    FORMAT_MESSAGE_IGNORE_INSERTS|FORMAT_MESSAGE_MAX_WIDTH_MASK,
	    hPDH, id, 0, sysMsgSpace+chars, (512-chars), 0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}


static int
PerfMonObjCmd (ClientData cData, Tcl_Interp *interp, int objc,
	     struct Tcl_Obj * CONST objv[])
{
    HQUERY hQuery;
    HCOUNTER *pCounterHandle;
    PDH_STATUS pdhStatus;
    DWORD ctrType;
    PDH_FMT_COUNTERVALUE fmtValue;

    pCounterHandle = (HCOUNTER *) ckalloc(sizeof(HCOUNTER));
    
    if ((pdhStatus = pdhProcs.PdhOpenQueryProc(0, 0, &hQuery))
	    != NO_ERROR) {
       	goto pdhError;
    }

    if ((pdhStatus = pdhProcs.PdhAddCounterProc(hQuery,
	    Tcl_GetString(objv[1]), 0, pCounterHandle))
	    != NO_ERROR) {
       	goto pdhError;
    }

    if ((pdhStatus = pdhProcs.PdhCollectQueryDataProc(hQuery))
	    != NO_ERROR) {
       	goto pdhError;
    }

    if (gotWideInt) {

	/* Get the 64-bit size */
	if ((pdhStatus = pdhProcs.PdhGetFormattedCounterValueProc(
		*pCounterHandle, PDH_FMT_LARGE|PDH_FMT_NOSCALE, &ctrType,
		&fmtValue)) != NO_ERROR) {
       	    goto pdhError;
	}

	/* Set a 64-bit native object.  Signed behavior is unknown,
	 * but at 2^63, who cares... */
	Tcl_SetObjResult(interp, Tcl_NewWideIntObj(
		fmtValue.largeValue));

    } else {

	/* Get the 32-bit size */
	if ((pdhStatus = pdhProcs.PdhGetFormattedCounterValueProc(
		*pCounterHandle, PDH_FMT_LONG|PDH_FMT_NOSCALE, &ctrType,
		&fmtValue)) != NO_ERROR) {
       	    goto pdhError;
	}

	/* 2^31 max possible. signed behavior is unknown. */
	Tcl_SetObjResult(interp, Tcl_NewIntObj(fmtValue.longValue));
    }

    if ((pdhStatus = pdhProcs.PdhCloseQueryProc(hQuery)) != NO_ERROR) {
	goto pdhError;
    }

    ckfree((char *)pCounterHandle);
    return TCL_OK;

pdhError:
    ckfree((char *)pCounterHandle);
    Tcl_SetObjResult(interp, GetPdhMsgObj(pdhStatus));
    return TCL_ERROR;
}


static void
Perfmon_ExitProc(ClientData clientData)
{
    /* be festideous */
    FreeLibrary(hPDH);
}


int
Perfmon_Init(Tcl_Interp *interp)
{
    Tcl_MutexLock(&initLock);

    if (!initialized) {
	OSVERSIONINFO os;
	int major, minor, patchLevel, type;

	initialized = 1;
	os.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&os);

	if (os.dwPlatformId == VER_PLATFORM_WIN32_NT) {
	    if ((hPDH = LoadLibrary("pdh.dll")) == 0L) {
		Tcl_SetObjResult(interp, Tcl_NewStringObj(
			"Can't load pdh.dll",-1));
		return TCL_ERROR;
	    }
	    usePerfMon = 1;
	    (void *) pdhProcs.PdhOpenQueryProc =
		    GetProcAddress(hPDH, "PdhOpenQueryA");
	    (void *) pdhProcs.PdhAddCounterProc =
		    GetProcAddress(hPDH, "PdhAddCounterA");
	    (void *) pdhProcs.PdhCollectQueryDataProc =
		    GetProcAddress(hPDH, "PdhCollectQueryData");
	    (void *) pdhProcs.PdhGetFormattedCounterValueProc =
		    GetProcAddress(hPDH, "PdhGetFormattedCounterValue");
	    (void *) pdhProcs.PdhCloseQueryProc =
		    GetProcAddress(hPDH, "PdhCloseQuery");
	    Tcl_CreateExitHandler(Perfmon_ExitProc, 0L);
	}
	Tcl_GetVersion(&major, &minor, &patchLevel, &type);
	/* Check for the Tcl_WideInt type by core version. */
	if (major > 8 || (major == 8 && minor > 4) || (major == 8 &&
		minor == 4 && ((type == TCL_ALPHA_RELEASE &&
		patchLevel >= 4) || type > TCL_ALPHA_RELEASE)))
	{
	    /* use the new Tcl_WideInt type when available */
	    gotWideInt = 1;
	}
    }

    Tcl_MutexUnlock(&initLock);

    /* no pdh.dll, no command, simple. */
    if (usePerfMon) {
	Tcl_CreateObjCommand(interp, "winutils::perfmon", PerfMonObjCmd, 0L,
	    0L);
    }
    return TCL_OK;
}
