/* ----------------------------------------------------------------------------
 * ProcessExtras.cpp --
 *
 *	Commands related to process management.
 *
 * ----------------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this.  See license.txt
 * for more detail.
 *
 * ----------------------------------------------------------------------------
 * RCS: @(#) $Id: ProcessExtras.cpp,v 1.1 2003/04/06 23:27:49 davygrvy Exp $
 * ----------------------------------------------------------------------------
 */

#include "Mcl/include/CMcl.h"
#ifdef _DEBUG
#pragma comment(lib,"mcld.lib")
#else
#pragma comment(lib,"mcl.lib")
#endif


#include "tclInt.h"	// for the namespace and callframe APIs
#ifdef _MSC_VER
    /* Only do this when MSVC++ is compiling us. */
#   ifdef USE_TCL_STUBS
	/* Mark this .obj as needing tcl's Stubs library. */
#	pragma comment(lib, "tclstub" \
		STRINGIFY(JOIN(TCL_MAJOR_VERSION,TCL_MINOR_VERSION)) ".lib")
#	if !defined(_MT) || !defined(_DLL) || defined(_DEBUG)
	    /* This fixes a bug with how the Stubs library was compiled.
	     * The requirement for msvcrt.lib from tclstubXX.lib should
	     * be removed. */
#	    pragma comment(linker, "-nodefaultlib:msvcrt.lib")
#	endif
#   else
	/* Mark this .obj needing the import library */
#	pragma comment(lib, "tcl" \
		STRINGIFY(JOIN(TCL_MAJOR_VERSION,TCL_MINOR_VERSION)) ".lib")
#   endif
#endif


static Tcl_Obj *
GetSysMsgObj(DWORD id)
{
    int chars;
    char sysMsgSpace[512];

    chars = wsprintf(sysMsgSpace, "[%u] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_IGNORE_INSERTS |
	    FORMAT_MESSAGE_MAX_WIDTH_MASK, 0L, id, 0, sysMsgSpace+chars, (512-chars),
	    0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}



// For CreateToolhelp32Snapshot()
#include <tlhelp32.h>


// Protos for our new commands
static Tcl_ObjCmdProc CrashObjCmd;
static Tcl_ObjCmdProc SendConSigObjCmd;
static Tcl_ObjCmdProc WaitPidObjCmd;
static Tcl_ObjCmdProc ProcessesObjCmd;

int
CrashObjCmd (ClientData clientData, Tcl_Interp *interp, int objc,
	struct Tcl_Obj * const objv[])
{
    DWORD pid;
    BOOL ok;
    HANDLE child;
    int exitCode = 7;  /* I saw this used in TclX... dunno why 7 was used. */

    if (objc < 2 || objc > 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "pid ?exitCode?");
    }
    if (Tcl_GetLongFromObj(interp, objv[1], (long *)&pid) != TCL_OK) {
	return TCL_ERROR;
    }
    if (objc == 3) {
	if (Tcl_GetIntFromObj(interp, objv[2], &exitCode) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if (exitCode < 0 || exitCode > 127) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("exitcode out-of-range (0-127).", -1));
	return TCL_ERROR;
    }

    child = OpenProcess(PROCESS_TERMINATE, FALSE, pid);
    if (!child) {
	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	return TCL_ERROR;
    }

    ok = TerminateProcess(child, exitCode);
    CloseHandle(child);
    if (!ok) {
	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	return TCL_ERROR;
    }

    return TCL_OK;
}


int
SendConSigObjCmd (ClientData clientData, Tcl_Interp *interp, int objc,
	struct Tcl_Obj * const objv[])
{
    static char *signals[] = {"ctrl+c", "ctrl+break", 0L};
    enum signals {SIG_CTRL_C, SIG_CTRL_BREAK};
    int signal;
    DWORD pid, event;
    BOOL ok;

    if (objc != 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "pid signal");
	return TCL_ERROR;
    }

    if (Tcl_GetLongFromObj(interp, objv[1], (long *)&pid) != TCL_OK) {
	return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[2], (const char **)signals, "signal", 0,
	    &signal) != TCL_OK) {
	goto end;
    }
    switch ((enum signals) signal) {
	case SIG_CTRL_C:
	    event = CTRL_C_EVENT; break;
	case SIG_CTRL_BREAK:
	    event = CTRL_BREAK_EVENT; break;
    }

    ok = GenerateConsoleCtrlEvent(event, pid);

    if (!ok) {
	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	return TCL_ERROR;
    }

end:
    return TCL_ERROR;
}

int
WaitPidObjCmd (ClientData cdata, Tcl_Interp *interp, int objc,
    Tcl_Obj *CONST objv[])
{
    DWORD pid, timeout, waitStatus, exitCode;
    HANDLE child;

    if (objc < 2 || objc > 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "pid ?timeout-milliseconds?");
    }
    if (Tcl_GetLongFromObj(interp, objv[1], (long *)&pid) != TCL_OK) {
	return TCL_ERROR;
    }
    if (objc == 3) {
	if (Tcl_GetLongFromObj(interp, objv[2], (long *)&timeout) != TCL_OK) {
	    return TCL_ERROR;
	}
    } else {
	timeout = INFINITE;
    }


    child = OpenProcess(SYNCHRONIZE, FALSE, pid);
    if (!child) {
	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	return TCL_ERROR;
    }

    waitStatus = WaitForSingleObject(child, timeout);

    switch (waitStatus) {
    case WAIT_FAILED:
	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	return TCL_ERROR;
    case WAIT_TIMEOUT:
	break;
    case WAIT_OBJECT_0:
	// Good, get and return the exit status.
	GetExitCodeProcess(child, &exitCode);
	Tcl_SetObjResult(interp, Tcl_NewLongObj(exitCode));
    }

    return TCL_OK;
}

int
ProcessesObjCmd(ClientData cdata, Tcl_Interp *interp, int objc,
    Tcl_Obj *CONST objv[])
{
    Tcl_Obj *resultObj = Tcl_GetObjResult(interp);
    HANDLE hSnapShot;
    PROCESSENTRY32 procentry;
    BOOL bFlag;

    // (FROM Q175030) Get a handle to a Toolhelp snapshot of all processes.
    hSnapShot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
    if (hSnapShot == INVALID_HANDLE_VALUE) {
	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
        return TCL_ERROR;
    }

    // (ADAPTED FROM Q175030) Get the first process' information.
    procentry.dwSize = sizeof(PROCESSENTRY32);
    bFlag = Process32First(hSnapShot, &procentry);
    if (bFlag == FALSE) {
	// No processes running??
        CloseHandle(hSnapShot);
        return TCL_OK;
    }

    // Loop until we've gotten all processes
    while (bFlag) {
	Tcl_Obj *list[2];

        // Add a new line of text for this process entry
        list[0] = Tcl_NewStringObj(procentry.szExeFile, -1);
        list[1] = Tcl_NewIntObj(procentry.th32ProcessID);
	Tcl_ListObjAppendElement(interp, resultObj, Tcl_NewListObj(2, list));

        // Get next process
        bFlag = Process32Next(hSnapShot, &procentry);
    }

    // Close the snapshot handle
    CloseHandle(hSnapShot);

    return TCL_OK;
}

extern "C" int
ProcessExtras_Init (Tcl_Interp *interp)
{
    Tcl_CreateObjCommand(interp, "winutils::crash", CrashObjCmd, 0L, 0L);
    Tcl_CreateObjCommand(interp, "winutils::sendconsig", SendConSigObjCmd, 0L, 0L);
    Tcl_CreateObjCommand(interp, "winutils::processes", ProcessesObjCmd, 0L, 0L);
    Tcl_CreateObjCommand(interp, "winutils::waitpid", WaitPidObjCmd, 0L, 0L);
    //Tcl_CreateObjCommand(interp, "winutils::waitpidcallback", WaitPidCallbackObjCmd, 0L, 0L);
    return TCL_OK;
}

