/*************************************************************
 ** Shell.c  --  (was) part of the Falcon DLL Extension v1.60
 **
 ** Abstract: This file contains the FALC_shell command.
 **
 ** Copyright: see 'main.cpp'
 **
 ** Author: David Gravereaux  mailto:davygrvy@pobox.com
 ************************************************************/

#include "tcl.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>

#ifdef _MSC_VER
    /* ShellExecute is found in this OS library. */
#   pragma comment(lib, "shell32.lib")
#endif


/*
 *-----------------------------------------------------------------------------
 *
 * ShellObjCmd --
 *
 *	a non-blocking way to start external apps using WinAPI's
 *	ShellExecuteEx().  I found the [exec] command to be a pain when I
 *	didn't want the exit code returned or redirection.
 *
 * Results:
 *	nothing or an error
 *
 * Side effects:
 *	Launched applications won't exit when Tcl does.
 *
 *-----------------------------------------------------------------------------
 */

int
ShellObjCmd (ClientData cData, Tcl_Interp *interp, int objc,
	     struct Tcl_Obj * CONST objv[])
{
    char *arg;
    int j, done = 0;
    BOOL doWait = FALSE;
    Tcl_Obj *params = 0L, *directory = 0L, *file = 0L;
    SHELLEXECUTEINFO sei;
    static char *sstyles[] = { "hide", "maximum", "minimum", "normal", NULL };
    enum sstyles { SHOW_HIDE, SHOW_MAX, SHOW_MIN, SHOW_NORM };
    int sstyle;
    static char *options[] = { "-verb", "-directory", "-show", "-wait", NULL };
    enum options { SHELL_VERB, SHELL_DIR, SHELL_SHOW, SHELL_WAIT };
    int option;
    char *p;

    ZeroMemory(&sei, sizeof(SHELLEXECUTEINFO));
    sei.cbSize = sizeof(SHELLEXECUTEINFO);
    sei.fMask = SEE_MASK_FLAG_NO_UI | SEE_MASK_NOCLOSEPROCESS | SEE_MASK_FLAG_DDEWAIT;
    sei.nShow = SW_SHOWNORMAL;

    /* Decode args */
    if (objc > 1) {
	for (j = 1; (j < objc) && !done; j++) {
	    arg = Tcl_GetString(objv[j]);
	    if (*arg == '-') {
		if (Tcl_GetIndexFromObj(interp, objv[j], options,
			"option", 0, &option) != TCL_OK) {
		    return TCL_ERROR;
		}
		switch ((enum options) option) {
		case SHELL_VERB:  /* verb */
		    if (objc>j+1) {
			sei.lpVerb = Tcl_GetString(objv[++j]);
		    } else {
			Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				"The -verb option requires a parameter.  ", 0L);
			goto Usage;
		    }
		    break;

		case SHELL_DIR: /* directory */
		    if (objc>j+1) {
			directory = Tcl_DuplicateObj(objv[++j]);
			for (p = Tcl_GetString(directory); *p != '\0'; p++) {
			    if (*p == '/') *p = '\\';
			}
			sei.lpDirectory = Tcl_GetString(directory);
		    } else {
			Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				"The -directory option requires a parameter.  ", 0L);
			goto Usage;
		    }
		    break;

		case SHELL_SHOW: /* show style */
		    if (objc>j+1) {
			if (Tcl_GetIndexFromObj(interp, objv[++j], sstyles,
				"ShowStyle", 0, &sstyle) != TCL_OK) {
			    return TCL_ERROR;
			}

			switch ((enum sstyles) sstyle) {
			case SHOW_HIDE:
			    sei.nShow = SW_HIDE; break;
			case SHOW_MAX:
			    sei.nShow = SW_SHOWMAXIMIZED; break;
			case SHOW_MIN:
			    sei.nShow = SW_SHOWMINNOACTIVE; break;
			case SHOW_NORM:
			    sei.nShow = SW_SHOWNORMAL;
			}
			break;
		    } else {
			Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
				"The -show option requires a parameter.  ", 0L);
			goto Usage;
		    }

		case SHELL_WAIT: /* wait */
		    doWait = TRUE;
		}
	    } else {
		done = 1;
		/* back it up one. */
		--j;
	    }
	}

	if (objc == j) {
	    Tcl_SetObjResult(interp, Tcl_NewStringObj("Nothing to execute.", -1));
	    return TCL_ERROR;
	}

	file = Tcl_DuplicateObj(objv[j++]);
	for (p = Tcl_GetString(file); *p != '\0'; p++) {
	    if (*p == '/') *p = '\\';
	}
	sei.lpFile = Tcl_GetString(file);

	/* everything left are parameters */
	if (objc > j) {
	    params = Tcl_ConcatObj(objc-j, &objv[j]);
	    sei.lpParameters = Tcl_GetString(params);
	}
    } else {
Usage:
	Tcl_WrongNumArgs(interp, 1, objv,
		"[-v <verb>] [-d <DefDirectory>] [-s <ShowStyle>] [-w] <file> [<param> ...]");
	return TCL_ERROR;
    }

    ShellExecuteEx(&sei);

    if (params) Tcl_DecrRefCount(params);
    if (directory) Tcl_DecrRefCount(directory);
    if (file) Tcl_DecrRefCount(file);

    if((DWORD)sei.hInstApp <= 32) {
	const char *msg;

	switch ((DWORD)sei.hInstApp) {
	    case SE_ERR_FNF:
		msg = "File not found."; break;
	    case SE_ERR_PNF:
		msg = "Path not found."; break;
	    case SE_ERR_ACCESSDENIED:
		msg = "Access denied."; break;
	    case SE_ERR_OOM:
		msg = "Out of memory."; break;
	    case SE_ERR_DLLNOTFOUND:
		msg = "Dynamic-link library not found."; break;
	    case SE_ERR_SHARE:
		msg = "Cannot share an open file."; break;
	    case SE_ERR_ASSOCINCOMPLETE:
		msg = "File association information is not complete."; break;
	    case SE_ERR_DDETIMEOUT:
		msg = "DDE operation has timed out."; break;
	    case SE_ERR_DDEFAIL:
		msg = "DDE operation failed."; break;
	    case SE_ERR_DDEBUSY:
		msg = "DDE operation is busy."; break;
	    case SE_ERR_NOASSOC:
		msg = "File association is not available."; break;
	    default:
		msg = "Unknown error."; break;
	}

	Tcl_SetObjResult(interp, Tcl_NewStringObj(msg, -1));
	return TCL_ERROR;
    }

    if (sei.hProcess != 0L) {
	if (doWait) {
	    /* no fancy error checking needed :) */
	    WaitForSingleObject(sei.hProcess, INFINITE);
	    GetExitCodeProcess(sei.hProcess, &doWait);
	    Tcl_SetObjResult(interp, Tcl_NewLongObj(doWait));
	}
	CloseHandle(sei.hProcess);
    }

    return TCL_OK;
}

int
Shell_Init(Tcl_Interp *interp)
{
    Tcl_CreateObjCommand(interp, "winutils::shell", ShellObjCmd, 0L, 0L);
    return TCL_OK;
}