/* ----------------------------------------------------------------------------
 * WinHelp.c --
 *
 *	The winhelp command.  Code borrowed from TclNT (a berkeley project).
 *
 * ----------------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this.  See license.txt
 * for more detail.
 *
 * ----------------------------------------------------------------------------
 * RCS: @(#) $Id: WinHelp.c,v 1.4 2002/04/21 06:19:37 davygrvy Exp $
 * ----------------------------------------------------------------------------
 */

#include "tcl.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#ifndef HELP_FINDER
#   define HELP_FINDER	    0x000B
#endif


static Tcl_Obj *GetSysMsgObj(DWORD id)
{
    int chars;
    char sysMsgSpace[512];

    chars = wsprintf(sysMsgSpace, "[%u] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
	    FORMAT_MESSAGE_IGNORE_INSERTS|FORMAT_MESSAGE_MAX_WIDTH_MASK,
	    0L, id, 0, sysMsgSpace+chars, (512-chars), 0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}

/*
 *--------------------------------------------------------------
 *
 * WinHelpObjCmd --
 *	Provides an interface to the Windows help system.
 *
 * Results:
 *	A standard TCL result.
 *
 * Usage: winutils::winhelp 
 *
 * Side effects:
 *	Will run winhlp32.exe if everything works correctly.
 *
 *--------------------------------------------------------------
 */
static int
WinHelpObjCmd (ClientData clientData, Tcl_Interp *interp, int objc,
	       struct Tcl_Obj * CONST objv[])
{
    HWND hWnd = 0L;
    char *filename;
    int num;
    DWORD dwData;
    BOOL ok;
    UINT fuCommand;
    static char *commands[] = {"command", "contents", "context", "contextpopup",
	    "forcefile", "helponhelp", "index", "key", "partialkey", "quit",
	    "setcontents", "setindex", "finder", 0L};
    enum commands {WH_COMMAND, WH_CONTENTS, WH_CONTEXT, WH_CONTEXTPOPUP,
	    WH_FORCEFILE, WH_HONH, WH_INDEX, WH_KEY, WH_PARTKEY, WH_QUIT,
	    WH_SETCONTENTS, WH_SETINDEX, WH_FINDER};
    int command;

    if (objc < 3 || objc > 5) {
	Tcl_WrongNumArgs(interp, 1, objv, "<filename> <optionname> ...\"");
	return TCL_ERROR;
    }

    filename = Tcl_GetString(objv[1]);

    if (Tcl_GetIndexFromObj(interp, objv[2], commands,
	    "command", 0, &command) != TCL_OK) {
	return TCL_ERROR;
    }
    switch ((enum commands) command) {
    case WH_COMMAND:
	fuCommand = HELP_COMMAND;
	dwData = (DWORD) Tcl_GetString(objv[3]);
	break;

    case WH_CONTENTS:
	if (Tcl_GetIntFromObj(interp, objv[3], &num) != TCL_OK) {
	    return TCL_ERROR;
	}
	fuCommand = HELP_CONTENTS;
	dwData = 0L;
	break;

    case WH_CONTEXT:
	if (Tcl_GetIntFromObj(interp, objv[3], &num) != TCL_OK) {
	    return TCL_ERROR;
	}
	fuCommand = HELP_CONTEXT;
	dwData = (DWORD) num;
	break;

    case WH_CONTEXTPOPUP:
	if (Tcl_GetIntFromObj(interp, objv[3], &num) != TCL_OK) {
	    return TCL_ERROR;
	}
	fuCommand = HELP_CONTEXTPOPUP;
	dwData = (DWORD) num;
	break;

    case WH_FORCEFILE:
	fuCommand = HELP_FORCEFILE;
	dwData = 0L;
	break;

    case WH_HONH:
	fuCommand = HELP_HELPONHELP;
	dwData = 0L;
	break;

    case WH_INDEX:
	fuCommand = HELP_INDEX;
	dwData = 0L;
	break;

    case WH_KEY:
	fuCommand = HELP_KEY;
	dwData = (DWORD) Tcl_GetString(objv[3]);
	break;

    case WH_PARTKEY:
	fuCommand = HELP_PARTIALKEY;
	dwData = (DWORD) Tcl_GetString(objv[3]);
	break;

    case WH_QUIT:
	fuCommand = HELP_QUIT;
	dwData = 0L;
	break;

    case WH_SETCONTENTS:
	if (Tcl_GetIntFromObj(interp, objv[3], &num) != TCL_OK) {
	    return TCL_ERROR;
	}
	fuCommand = HELP_SETCONTENTS;
	dwData = (DWORD) num;
	break;

    case WH_SETINDEX:
	if (Tcl_GetIntFromObj(interp, objv[3], &num) != TCL_OK) {
	    return TCL_ERROR;
	}
	fuCommand = HELP_SETINDEX;
	dwData = (DWORD) num;
	break;

    case WH_FINDER:
	fuCommand = HELP_FINDER;
	dwData = (DWORD) 0;
	break;
    }

    ok = WinHelp(hWnd, filename, fuCommand, dwData);

    if (!ok) {
	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	return TCL_ERROR;
    }

    return TCL_OK;
}

int
Winhelp_Init(Tcl_Interp *interp)
{
    Tcl_CreateObjCommand(interp, "winutils::winhelp", WinHelpObjCmd, 0L, 0L);
    return TCL_OK;
}