/* ---------------------------------------------------------------------
 * WinSockMisc.c --
 *
 *	Misc routines related to winsock.
 *
 * ---------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this code.  See
 * license.txt for more detail.
 *
 * ---------------------------------------------------------------------
 * RCS: @(#) $Id: WinSockMisc.c,v 1.2 2002/05/22 02:30:17 davygrvy Exp $
 * ---------------------------------------------------------------------
 */

#include "tcl.h"
#define _WIN32_WINNT 0x0500        /* Enables NT5 special features */
#define WIN32_LEAN_AND_MEAN
/* Forcefully prevents inclusion of winsock.h */
#define _WINSOCKAPI_
#include <windows.h>
/* Enable typedefs for run-time linking thru code with LoadLibrary() */
#define INCL_WINSOCK_API_TYPEDEFS 1
/* winsock(2).h should check for the intrinsic _WIN32, but doesn't. */
#ifndef WIN32
#define WIN32
#endif
#include <winsock2.h>
#if !defined(_WS2TCPIP_H_)
# include <ws2tcpip.h>
#endif

static struct {
    HMODULE hInstance;    /* Handle to WinSock library. */
    int ver;
    LPFN_GETSERVBYNAME getservbyname;
    LPFN_GETSERVBYPORT getservbyport;
    LPFN_HTONS htons;
    LPFN_HTONL htonl;
    LPFN_INET_ADDR inet_addr;
    LPFN_INET_NTOA inet_ntoa;
    LPFN_NTOHL ntohl;
    LPFN_NTOHS ntohs;
    LPFN_WSASTARTUP WSAStartup;
    LPFN_WSACLEANUP WSACleanup;
    LPFN_WSAGETLASTERROR WSAGetLastError;
    LPFN_WSASTRINGTOADDRESSA WSAStringToAddress;
} winSock;

TCL_DECLARE_MUTEX(initLock)
static int initialized = 0;
static int sockGood = 1;

/*
 *----------------------------------------------------------------------
 *
 * InitSockets --
 *
 *	Initializes WinSock when the DLL is first loaded.
 *
 * Results:
 *	TRUE for success, or FALSE for failure.
 *
 * Side effects:
 *	A winsock DLL is now sitting in our address space.
 *
 *----------------------------------------------------------------------
 */
BOOL InitSockets (void)
{
    WORD wVersionRequired;
    WSADATA wsaData;

    /* try winsock2 first */
    winSock.hInstance = LoadLibrary("ws2_32.dll");
    wVersionRequired = MAKEWORD(2,2);
    winSock.ver = 2;

    if (winSock.hInstance == 0L) {
	winSock.hInstance = LoadLibrary("wsock32.dll");
	wVersionRequired = MAKEWORD(1,1);
	winSock.ver = 1;
    }

    /*
     * Initialize the function table.
     */

    if (winSock.hInstance == 0L) return FALSE;
    
    (void *) winSock.getservbyname   = GetProcAddress(winSock.hInstance, "getservbyname");
    (void *) winSock.getservbyport   = GetProcAddress(winSock.hInstance, "getservbyport");
    (void *) winSock.htonl           = GetProcAddress(winSock.hInstance, "htonl");
    (void *) winSock.htons           = GetProcAddress(winSock.hInstance, "htons");
    (void *) winSock.inet_addr       = GetProcAddress(winSock.hInstance, "inet_addr");
    (void *) winSock.inet_ntoa       = GetProcAddress(winSock.hInstance, "inet_ntoa");
    (void *) winSock.ntohl           = GetProcAddress(winSock.hInstance, "ntohl");
    (void *) winSock.ntohs           = GetProcAddress(winSock.hInstance, "ntohs");
    (void *) winSock.WSAStartup      = GetProcAddress(winSock.hInstance, "WSAStartup");
    (void *) winSock.WSACleanup      = GetProcAddress(winSock.hInstance, "WSACleanup");
    (void *) winSock.WSAGetLastError = GetProcAddress(winSock.hInstance, "WSAGetLastError");
    (void *) winSock.WSAStringToAddress = GetProcAddress(winSock.hInstance, "WSAStringToAddressA");

    /*
     * Now check that all fields are properly initialized. If not, return
     * FALSE to indicate that we failed to initialize properly.
     */

    if ((winSock.getservbyname == 0L) ||
	(winSock.getservbyport == 0L) ||
	(winSock.htonl == 0L) ||
	(winSock.htons == 0L) ||
	(winSock.inet_addr == 0L) ||
	(winSock.inet_ntoa == 0L) ||
	(winSock.ntohl == 0L) ||
	(winSock.ntohs == 0L) ||
	(winSock.WSAStartup == 0L) ||
	(winSock.WSACleanup == 0L) ||
	(winSock.WSAGetLastError == 0L))
    {
	goto unloadLibrary;
    }
    
    /*
    * Initialize the winsock library and check the version number.
    */
    
    if (winSock.WSAStartup(wVersionRequired, &wsaData) != 0) {
	goto unloadLibrary;
    }
    
    if (wsaData.wVersion != wVersionRequired) {
	winSock.WSACleanup();
	goto unloadLibrary;
    }
    
    return TRUE;
    
unloadLibrary:
    FreeLibrary(winSock.hInstance);
    winSock.hInstance = 0L;
    return FALSE;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdIsip --
 *
 *  checks if the given string is an IP address
 *
 * Results:
 *  returns 1/0
 *
 * Side effects:
 *  The inet_addr() function doesn't seem to mind if you leave out 
 *  the "." (dots) in the string. I think that's strange behavior.
 *
 *----------------------------------------------------------------------
 */
int
Isip4ObjCmd (ClientData cData, Tcl_Interp *interp, int objc,
	    struct Tcl_Obj * CONST objv[])
{
    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "<text> ; returns true if "
		"<text> is an IPv4 address in dotted 4 form. Or false, "
		"if not an IPv4.");
	return TCL_ERROR;
    }

    if (winSock.inet_addr(Tcl_GetString(objv[1])) == INADDR_NONE) {
	Tcl_SetObjResult(interp, Tcl_NewBooleanObj(0));
	return TCL_OK;
    }

    Tcl_SetObjResult(interp, Tcl_NewBooleanObj(1));
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Ip2LongCmd --
 *
 *	returns the network long IP address from the given text
 *
 * Results:
 *	returns the network-order address in long form, or zero for an
 *	error.
 *
 * Side effects:
 *	The inet_addr() function doesn't seem to mind if you leave out 
 *	the "." (dots) in the string. I think that's strange behavior.
 *
 *----------------------------------------------------------------------
 */
int
Ip2LongCmd (ClientData cData, Tcl_Interp *interp, int objc,
	    struct Tcl_Obj * CONST objv[])
{
    unsigned long addr=0;

    if (objc < 2){
	Tcl_WrongNumArgs(interp, 1, objv, "<text> ; returns the long IP "
		"in network-order, if <text> is an IP address in dotted "
		"4 form. Or zero, if not an IP.");
	return TCL_ERROR;
    }

    if((addr = winSock.inet_addr(Tcl_GetString(objv[1]))) != INADDR_NONE) {
	if ((int)cData == 1) {
	    Tcl_SetObjResult(interp, Tcl_NewLongObj(winSock.htonl(addr)));
	} else {
	    Tcl_SetObjResult(interp, Tcl_NewLongObj(addr));
	}
	return TCL_OK;
    } else {
	Tcl_SetObjResult(interp, Tcl_NewLongObj(0));
	return TCL_OK;
    };
}

/*
 *----------------------------------------------------------------------
 *
 * CmdLong2Ip --
 *
 *  returns a dotted4 IP from a network long IP number
 *
 * Results:
 *  see above
 *
 * Side effects:
 *  The inet_addr() function doesn't seem to mind if you leave out 
 *  the "." (dots) in the string. I think that's strange behavior.
 *
 *----------------------------------------------------------------------
 */
int
Long2IpCmd (ClientData cData, Tcl_Interp *interp, int objc,
	    struct Tcl_Obj * CONST objv[])
{
    IN_ADDR addr;
    unsigned long longAddr;
    char *dot4;

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "<text> ; returns the IP in "
		"dotted decimal from a long network-order IP address. "
		"Or zero, if not an long IP.");
	return TCL_ERROR;
    }

    if (Tcl_GetLongFromObj(interp, objv[1], &longAddr) != TCL_OK) {
	return TCL_OK;
    }

    if ((int)cData == 1) {
	addr.s_addr = longAddr;
    } else {
	addr.s_addr = winSock.htonl(longAddr);
    }

    if ((dot4 = winSock.inet_ntoa(addr)) != 0L) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj(dot4, -1));
	return TCL_OK;
    } else {
	Tcl_SetObjResult(interp, Tcl_NewLongObj(0));
	return TCL_OK;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CmdServ --
 *	Resolves a service name to a port number.
 *
 * Results:
 *	returns the port # from a service name
 *
 * Side effects:
 *	unknown.
 *
 *----------------------------------------------------------------------
 */
static int
GetPortByNameCmd (ClientData cData, Tcl_Interp *interp,int objc,
		  struct Tcl_Obj * CONST objv[])
{
    struct servent *result = 0L;
    int dummy;

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "<ServiceName>|<PortNum>");
	return TCL_ERROR;
    }

    if (Tcl_GetIntFromObj(0L, objv[1], &dummy) != TCL_ERROR) {
	Tcl_SetObjResult(interp, objv[1]);
    } else {
	result = winSock.getservbyname(Tcl_GetString(objv[1]),
		(const char *) cData);
	if (result != 0L) {
	    Tcl_SetObjResult(interp,
		    Tcl_NewIntObj(winSock.ntohs(result->s_port)));
	} else {
	    Tcl_SetObjResult(interp, Tcl_NewStringObj(
		    "Service name unknown.", -1));
	    return TCL_ERROR;
	}
    }

    return TCL_OK;
}

/*
 * ---------------------------------------------------------------------
 * The Tcl_PackageInitProc for this section.
 * ---------------------------------------------------------------------
 */

int
Winsockmisc_Init(Tcl_Interp *interp)
{
    Tcl_MutexLock(&initLock);
    if (!initialized) {
	initialized = 1;

	/*
	 * Sockets aren't available, so these commands can't be created.
	 * This isn't a fatal error.
	 */
	if (InitSockets() == FALSE) {
	    sockGood = 0;
	    Tcl_MutexUnlock(&initLock);
	    return TCL_OK;
	}
    }

    Tcl_MutexUnlock(&initLock);
    
    if (sockGood) {
	Tcl_CreateObjCommand(interp, "winutils::isip4", Isip4ObjCmd, 0L, 0L);
	Tcl_CreateObjCommand(interp, "winutils::ip2netlong", Ip2LongCmd, (ClientData)1L, 0L);
	Tcl_CreateObjCommand(interp, "winutils::ip2long", Ip2LongCmd, 0L, 0L);
	Tcl_CreateObjCommand(interp, "winutils::netlong2ip", Long2IpCmd, (ClientData)1L, 0L);
	Tcl_CreateObjCommand(interp, "winutils::long2ip", Long2IpCmd, 0L, 0L);
	Tcl_CreateObjCommand(interp, "winutils::tcpport", GetPortByNameCmd, "tcp", 0L);
	Tcl_CreateObjCommand(interp, "winutils::udpport", GetPortByNameCmd, "udp", 0L);
    }
    return TCL_OK;
}
