//  filename: winapi.cpp
//  author:   David Gravereaux ( mailto:davygrvy@pobox.com )
//  when:     9:17 PM 2/27/2001

#include <tcl.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN

// We need at least the Tcl_Obj interface that was started in 8.0
#if TCL_MAJOR_VERSION < 8
#   error "we need Tcl 8.0 or greater to build this"

// Check for Stubs compatibility when asked for it.
#elif defined(USE_TCL_STUBS) && TCL_MAJOR_VERSION == 8 && \
	(TCL_MINOR_VERSION == 0 || \
        (TCL_MINOR_VERSION == 1 && TCL_RELEASE_LEVEL != TCL_FINAL_RELEASE))
#   error "Stubs interface doesn't work in 8.0 and alpha/beta 8.1; only 8.1.0+"
#endif

#ifdef _MSC_VER
    // Only do this when MSVC++ is compiling us.
#   ifdef USE_TCL_STUBS
	// Mark this .obj as needing tcl's Stubs library.
#	pragma comment(lib, "tclstub" \
		STRINGIFY(JOIN(TCL_MAJOR_VERSION,TCL_MINOR_VERSION)) ".lib")
#	if !defined(_MT) || !defined(_DLL) || defined(_DEBUG)
	    // This fixes a bug with how the Stubs library was compiled.
	    // The requirement for msvcrt.lib from tclstubXX.lib should
	    // be removed.
#	    pragma comment(linker, "-nodefaultlib:msvcrt.lib")
#	endif
#   else
	// Mark this .obj needing the import library
#	pragma comment(lib, "tcl" \
		STRINGIFY(JOIN(TCL_MAJOR_VERSION,TCL_MINOR_VERSION)) ".lib")
#   endif
#   pragma comment (lib, "user32.lib")
#endif


//=============================================================================


class TclAdaptor
{
    typedef int (TclAdaptor::*TclAdaptorCmd)(int, Tcl_Obj * CONST []);
    
    // used by all the Tcl commands for the demultiplexor
    // by storing its member function location in the
    // clientdata on command creation.
    //
    typedef struct {
	TclAdaptor *ext;
	TclAdaptorCmd cmd;
    } CMDINFO, *LPCMDINFO;

protected:
    explicit TclAdaptor(Tcl_Interp *interp)
	: m_interp(interp)
    {
	Tcl_CallWhenDeleted(interp, InterpDeleting, this);
	Tcl_CreateExitHandler(Exiting, this);
    }

    ~TclAdaptor()
    {
	Tcl_DeleteExitHandler(Exiting, this);
    }

    inline ClientData
    CmdInfo (TclAdaptorCmd cmd)
    {
	LPCMDINFO mplexData = new CMDINFO;
	mplexData->ext = this;
	mplexData->cmd = cmd;
	return static_cast<ClientData>(mplexData);
    }

    char *FQNS(const char *ns, const char *cmd)
    {
	strcpy(nsCmdBuff, ns);
	strcat(nsCmdBuff, "::");
	strcat(nsCmdBuff, cmd);
	return nsCmdBuff;
    }
    
    Tcl_Interp * m_interp;

private:
    friend int
	CmdCallback(ClientData, Tcl_Interp *, int, struct Tcl_Obj * CONST []);
    friend void CmdDelete(ClientData);
    friend Tcl_InterpDeleteProc InterpDeleting;
    friend Tcl_ExitProc Exiting;
    char nsCmdBuff[120];
};

static void
InterpDeleting (ClientData clientData, Tcl_Interp *)
{
    // Add clean-up code here that references the interp.
    //
    delete static_cast<TclAdaptor *>(clientData);
}

static void
Exiting (ClientData clientData)
{
    delete static_cast<TclAdaptor *>(clientData);
}


static int
CmdCallback (ClientData clientData, Tcl_Interp *,
	     int objc, struct Tcl_Obj * CONST objv[])
{
    TclAdaptor::LPCMDINFO mplex =
	    static_cast<TclAdaptor::LPCMDINFO>(clientData);
	    
    // We aleady know what the interp pointer is,
    // so don't bother with it. Call the member function
    // of the extension instance as saved in the
    // clientdata pointer.
    //
    // This is a demultiplexor or 'demux' for short.
    return ((mplex->ext) ->* (mplex->cmd)) (objc,objv);
}

static void
CmdDelete(ClientData clientData)
{
    delete static_cast<TclAdaptor::LPCMDINFO>(clientData);
}

//=============================================================================


/*
 * For determining the default button in the message box code
 */
#define CDLG_ABORT  1
#define CDLG_RETRY  2
#define CDLG_IGNORE 3
#define CDLG_OK	    4
#define CDLG_CANCEL 5
#define CDLG_YES    6
#define CDLG_NO	    7

class TclAskBox
    : protected TclAdaptor
{
public:
    explicit TclAskBox(Tcl_Interp *interp, const char *ns)
	: TclAdaptor(interp)
    {
	
    	Tcl_CreateObjCommand(interp, FQNS(ns,"askbox"), CmdCallback,
	    CmdInfo(reinterpret_cast<TclAdaptorCmd>(&TclAskBox::AskBoxCmd)),
	    CmdDelete);
    }
    
    int AskBoxCmd(int objc, struct Tcl_Obj * CONST objv[])
    {
	int indx;
	DWORD buttons = 0xFFFFFFFFL;
	DWORD icon = 0xFFFFFFFFL;
	DWORD dflt = 0xFFFFFFFFL;
	//DWORD help = 0;
	int len, retVal, error;
	Tcl_Obj *outMsgObj;
	char *title = "no title";
	char *s;
	char buff[33];
	HANDLE owner = NULL;
	MSGBOXPARAMS parms;
	//CMDLGCUSTDATA cust;


	if (objc < 2) {
    Usage:
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(m_interp),
		    "wrong # args: ", 0L);
    BadOpt:
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(m_interp),
		    "should be \"", Tcl_GetStringFromObj(objv[0], 0L),
		    /* " [-h <helpfile> <context>]", */
		    " [-t <Title>]",
		    " [-b abortretryignore|ok|okcancel|retrycancel|yesno|yesnocancel]",
		    " [-d abort|retry|ignore|ok|cancel|yes|no]",
		    " [-i exclamation|info|question|stop|0]",
		    " <message>", 0L);
	    return TCL_ERROR;
	}

	indx = 1;

	// decode args
	if (objc > 2) {
	    while (indx + 1 < objc) {
		char *opt = Tcl_GetStringFromObj(objv[indx], 0L);
		char *optVal;
		
		if (strcmp(opt, "-b") == 0) {
		    indx++;
		    if (indx == objc) {
			goto Usage;
		    }
		    if (buttons != 0xFFFFFFFFL) {
			Tcl_SetObjResult(m_interp, Tcl_NewStringObj(
				"error: specify the -b option only once",
				-1));
			return TCL_ERROR;
		    }
		    
		    optVal = Tcl_GetStringFromObj(objv[indx],0L);
		    
		    if (strcmp(optVal, "abortretryignore") == 0) {
			buttons = MB_ABORTRETRYIGNORE;
		    } else if (strcmp(optVal, "ok") == 0) {
			buttons = MB_OK;
		    } else if (strcmp(optVal, "okcancel") == 0) {
			buttons = MB_OKCANCEL;
		    } else if (strcmp(optVal, "retrycancel") == 0) {
			buttons = MB_RETRYCANCEL;
		    } else if (strcmp(optVal, "yesno") == 0) {
			buttons = MB_YESNO;
		    } else if (strcmp(optVal, "yesnocancel") == 0) {
			buttons = MB_YESNOCANCEL;
		    } else {
			Tcl_Obj *result = 
			    Tcl_NewStringObj("unknown button combination \"",
			    -1);
			Tcl_AppendStringsToObj(result, optVal, 
				"\" for -b option: should be one of ",
				"abortretryignore, ok, okcancel, ",
				"retrycancel, yesno, or yesnocancel", 0L);
			Tcl_SetObjResult(m_interp, result);
			return TCL_ERROR;
		    }

		} else if (strcmp(opt, "-d") == 0) {
		    indx++;
		    if (indx == objc) {
			goto Usage;
		    }

		    optVal = Tcl_GetStringFromObj(objv[indx],0L);
		    
		    if (strcmp(optVal, "ok") == 0) {
			dflt = CDLG_OK;
		    } else if (strcmp(optVal, "cancel") == 0) {
			dflt = CDLG_CANCEL;
		    } else if (strcmp(optVal, "yes") == 0) {
			dflt = CDLG_YES;
		    } else if (strcmp(optVal, "no") == 0) {
			dflt = CDLG_NO;
		    } else if (strcmp(optVal, "abort") == 0) {
			dflt = CDLG_ABORT;
		    } else if (strcmp(optVal, "retry") == 0) {
			dflt = CDLG_RETRY;
		    } else if (strcmp(optVal, "ignore") == 0) {
			dflt = CDLG_IGNORE;
		    } else if (optVal[0] == '0') {
			dflt = 0xFFFFFFFFL;
		    } else {
			Tcl_Obj *result =
				Tcl_NewStringObj("unknown button \"", -1);
			Tcl_AppendStringsToObj(result, optVal,
				"\" for -d option: should be one ",
				"of ok, cancel, yes, no, abort, retry, ",
				"or ignore", 0L);
			Tcl_SetObjResult(m_interp, result);
			return TCL_ERROR;
		    }

		} else if (strcmp(opt, "-i") == 0) {
		    indx++;
		    if (indx == objc) {
			goto Usage;
		    }

		    optVal = Tcl_GetStringFromObj(objv[indx],0L);
		    
		    if (strcmp(optVal, "exclamation") == 0) {
			icon = MB_ICONEXCLAMATION;
		    } else if (strcmp(optVal, "info") == 0) {
			icon = MB_ICONINFORMATION;
		    } else if (strcmp(optVal, "question") == 0) {
			icon = MB_ICONQUESTION;
		    } else if (strcmp(optVal, "stop") == 0) {
			icon = MB_ICONERROR;
		    } else if (optVal[0] == '0') {
			icon = 0;
		    } else {
			Tcl_Obj *result =
				Tcl_NewStringObj("unknown button \"", -1);
			Tcl_AppendStringsToObj(result, optVal, 
				"\" to -i option: should be one ",
				"of exclamation, info, question, stop",
				"or 0 to specify none", 0L);
			Tcl_SetObjResult(m_interp, result);
			return TCL_ERROR;
		    }

		} else if (strcmp(opt, "-t") == 0) {
		    indx++;
		    if (indx == objc) {
			goto Usage;
		    }
		    title = Tcl_GetStringFromObj(objv[indx],0L);
		
		} else {
		    Tcl_Obj *result =
			    Tcl_NewStringObj("unknown option \"", -1);
		    Tcl_AppendStringsToObj(result, opt, "\": ", 0L);
		    Tcl_AppendObjToObj(Tcl_GetObjResult(m_interp), result);
		    goto BadOpt;
		}

		indx++;
	    }
	}
	
	outMsgObj = Tcl_DuplicateObj(objv[indx++]);

	if (indx != objc) {
	    Tcl_WrongNumArgs(m_interp, 1, objv, "[<options>] <message>");
	    Tcl_DecrRefCount(outMsgObj);
	    return TCL_ERROR;
	}
	if (icon == 0xFFFFFFFFL) {
	    icon = 0;
	}
	if (buttons == 0xFFFFFFFFL) {
	    buttons = MB_OK;
	}

	error = 0;

	if (dflt == 0xFFFFFFFFL) {
	    dflt = MB_DEFBUTTON1;
	} else {
	    if (buttons == MB_ABORTRETRYIGNORE) {
		if (dflt == CDLG_ABORT) {
		    dflt = MB_DEFBUTTON1;
		} else if (dflt == CDLG_RETRY) {
		    dflt = MB_DEFBUTTON2;
		} else if (dflt == CDLG_IGNORE) {
		    dflt = MB_DEFBUTTON3;
		} else {
		    error = 1;
		}
	    
	    } else if (buttons == MB_OK) {
		if (dflt == CDLG_OK) {
		    dflt = MB_DEFBUTTON1;
		} else {
		    error = 1;
		}
	    
	    } else if (buttons == MB_OKCANCEL) {
		if (dflt == CDLG_OK) {
		    dflt = MB_DEFBUTTON1;
		} else if (dflt == CDLG_CANCEL) {
		    dflt = MB_DEFBUTTON2;
		} else {
		    error = 1;
		}

	    } else if (buttons == MB_RETRYCANCEL) {
		if (dflt == CDLG_RETRY) {
		    dflt = MB_DEFBUTTON1;
		} else if (dflt == CDLG_CANCEL) {
		    dflt = MB_DEFBUTTON2;
		} else {
		    error = 1;
		}

	    } else if (buttons == MB_YESNO) {
		if (dflt == CDLG_YES) {
		    dflt = MB_DEFBUTTON1;
		} else if (dflt == CDLG_NO) {
		    dflt = MB_DEFBUTTON2;
		} else {
		    error = 1;
		}

	    } else if (buttons == MB_YESNOCANCEL) {
		if (dflt == CDLG_YES) {
		    dflt = MB_DEFBUTTON1;
		} else if (dflt == CDLG_NO) {
		    dflt = MB_DEFBUTTON2;
		} else if (dflt == CDLG_CANCEL) {
		    dflt = MB_DEFBUTTON3;
		} else {
		    error = 1;
		}

	    } else {
		Tcl_SetObjResult(m_interp, Tcl_NewStringObj(
			"internal error", -1));
		Tcl_DecrRefCount(outMsgObj);
		return TCL_ERROR;
	    }
	}

	if (error) {
	    Tcl_SetObjResult(m_interp, Tcl_NewStringObj(
		    "-d option button does not exist in the specified button combination",
		    -1));
	    return TCL_ERROR;
	}
	

	// parse through the message and replace all \n's
	// with newlines.  Operate on the Tcl_Obj directly.
	//
	
	//    ip = text; op = textOut;
	//    while (*ip != 0) {
	//	if (*ip == '\\' && ip[1] == 'n') {
	//	    *op = '\n';
	//	    ip++;
	//	} else {
	//	    *op = *ip;
	//	}
	//	ip++; op++;
	//    }
	//    *op = 0;
	//    text = textOut;
	//}

	parms.hwndOwner = 0L;
	parms.cbSize = sizeof(MSGBOXPARAMS);
	parms.hInstance = 0L;
	parms.lpszIcon = 0L;
	parms.dwContextHelpId = 0L;
	parms.lpszText = Tcl_GetStringFromObj(outMsgObj, 0L);
	parms.lpszCaption = title;
	parms.lpfnMsgBoxCallback = 0L; //MsgBoxCallback;
	parms.dwStyle = buttons | dflt | icon | MB_SETFOREGROUND; 

	retVal = MessageBoxIndirect(&parms);

	switch (retVal) {
	case IDABORT:
	    s = "abort"; break;
	case IDCANCEL:
	    s = "cancel"; break;
	case IDIGNORE:
	    s = "ignore"; break;
	case IDNO:
	    s = "no"; break;
	case IDOK:
	    s = "ok"; break;
	case IDRETRY:
	    s = "retry"; break;
	case IDYES:
	    s = "yes"; break;
	default:
	    Tcl_SetObjResult(m_interp, Tcl_NewStringObj("out of memory", 0L));
	    Tcl_DecrRefCount(outMsgObj);
	    return TCL_ERROR;
	}

	Tcl_SetObjResult(m_interp, Tcl_NewStringObj(s, -1));
	Tcl_DecrRefCount(outMsgObj);
	return TCL_OK;
    }
};

/*

    int FindWindowCmd (int objc, struct Tcl_Obj * CONST objv[])
    {
	char *ptitle = NULL, *pclass = NULL;
	HWND rtn;
	int j;

	if (objc < 2 || objc > 4)
	{
	    Tcl_WrongNumArgs(m_interp, 1, objv,
		"?-class classname? titlebartext; returns the hWnd of the found window as a signed long value.");
	    return TCL_ERROR;
  	}

	// Decode args
	for (j=1; j<objc; j++)
	{
	    char *arg = Tcl_GetStringFromObj(objv[j], 0L);
	    if (*arg == '-')
	    {
		switch (*(arg + 1))
		{
		case 'c': // class
		    if (j+1 < objc)
		    {
			pclass = Tcl_GetStringFromObj(objv[++j], 0L);
		    } else {
			Tcl_SetObjResult(m_interp,
				Tcl_NewStringObj("-class option requires a value to follow it.", -1));
			return TCL_ERROR;
		    }
		    break;
		default:
		    Tcl_SetObjResult(m_interp, Tcl_NewStringObj("unknown option.", -1));
		    return TCL_ERROR;
		}
	    } else {
	      // no more options left, so ditch this loop
	      break;
	    }
	}

	if (j < objc)
	{
	    ptitle = Tcl_GetStringFromObj(objv[j], 0L);
	}

	rtn = FindWindow(pclass, ptitle);
	Tcl_SetObjResult(m_interp, Tcl_NewLongObj(reinterpret_cast<long>(rtn)));
	return TCL_OK;
    }

    Tcl_Interp *m_interp;
};
*/


class UIToolsExtension
    : virtual private TclAskBox
{
public:
    explicit UIToolsExtension(Tcl_Interp *interp)
	: TclAskBox(interp, "uitools")
    {
	Tcl_PkgProvide(interp, "uitools", "0.0.1");
    }
};

#undef TCL_STORAGE_CLASS
#define TCL_STORAGE_CLASS DLLEXPORT

EXTERN int
Uitools_Init (Tcl_Interp *interp)
{
#ifdef USE_TCL_STUBS
    if (Tcl_InitStubs(interp, TCL_VERSION, 0) == NULL) {
	return TCL_ERROR;
    }
#endif
    new UIToolsExtension(interp);
    return TCL_OK;
}

